from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers


class DummySearchResultModelSerializer(serializers.Serializer):
    """
    Empty serializer for Swagger views.
    """


class SearchFieldSerializer(serializers.Serializer):
    field_name = serializers.CharField(
        label=_(message='Field name'), read_only=True
    )
    label = serializers.CharField(
        label=_(message='Label'), read_only=True
    )


class SearchModelSerializer(serializers.Serializer):
    app_label = serializers.CharField(
        label=_(message='App label'), read_only=True
    )
    model_name = serializers.CharField(
        label=_(message='Model name'), read_only=True
    )
    pk = serializers.CharField(
        label=_(message='Primary key'), read_only=True
    )
    search_fields = SearchFieldSerializer(
        label=_(message='Search fields'), many=True, read_only=True
    )



from django.apps import apps
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from mayan.apps.templating.classes import Template

from .classes import EventType
from .literals import TEXT_UNKNOWN_EVENT_ID


def widget_event_actor_link(context, attribute=None):
    entry = context['object']

    ContentType = apps.get_model(
        app_label='contenttypes', model_name='ContentType'
    )

    if attribute:
        entry = getattr(entry, attribute)

    if entry.actor == entry.target:
        label = _(message='System')
        url = None
    else:
        label = entry.actor
        content_type = ContentType.objects.get_for_model(model=entry.actor)

        url = reverse(
            viewname='events:object_event_list', kwargs={
                'app_label': content_type.app_label,
                'model_name': content_type.model,
                'object_id': entry.actor.pk
            }
        )

    if url:
        return Template(
            template_string='<a href="{{ url }}">{{ label }}</a>'
        ).render(
            context={'label': entry.actor, 'url': url}
        )
    else:
        return label


def widget_event_type_link(context, attribute=None):
    entry = context['object']

    if attribute:
        entry = getattr(entry, attribute)

    try:
        event_type_label = EventType.get(id=entry.verb).label
    except KeyError:
        event_type_label = TEXT_UNKNOWN_EVENT_ID % entry.verb

    return mark_safe(
        s='<a href="{url}">{label}</a>'.format(
            url=reverse(
                viewname='events:verb_event_list',
                kwargs={'verb': entry.verb}
            ), label=event_type_label
        )
    )



from mayan.apps.dependencies.classes import PythonDependency

PythonDependency(
    module=__name__, name='gevent', version_string='==23.9.1'
)
PythonDependency(
    module=__name__, name='greenlet', version_string='==3.0.3'
)
PythonDependency(
    module=__name__, name='gunicorn', version_string='==21.2.0'
)
PythonDependency(
    module=__name__, name='sentry-sdk', version_string='==1.40.6'
)
PythonDependency(
    module=__name__, name='whitenoise', version_string='==6.6.0'
)



class ReadOnlyAdminMixin:
    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0053_auto_20200129_0926')
    ]

    operations = [
        migrations.CreateModel(
            name='TrashedDocument',
            fields=[],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': []
            },
            bases=('documents.document',),
        )
    ]



import datetime

from django.utils.translation import gettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_b, worker_c

from .settings import setting_workflow_state_escalation_check_interval

queue_workflows = CeleryQueue(
    label=_(message='Workflows'), name='workflows', worker=worker_b
)
queue_workflows_slow = CeleryQueue(
    label=_(message='Workflows slow'), name='workflows_slow', worker=worker_c
)

queue_workflows.add_task_type(
    label=_(message='Launch a workflow for a document'),
    dotted_path='mayan.apps.document_states.tasks.task_launch_workflow_for'
)
queue_workflows.add_task_type(
    label=_(message='Launch all workflows for a document'),
    dotted_path='mayan.apps.document_states.tasks.task_launch_all_workflow_for'
)
queue_workflows.add_task_type(
    label=_(message='Check a workflow instance for state escalation.'),
    dotted_path='mayan.apps.document_states.tasks.task_workflow_instance_check_escalation'
)

queue_workflows_slow.add_task_type(
    label=_(message='Check all workflow instances for state escalation.'),
    dotted_path='mayan.apps.document_states.tasks.task_workflow_instance_check_escalation_all',
    schedule=datetime.timedelta(
        seconds=setting_workflow_state_escalation_check_interval.value
    )
)
queue_workflows_slow.add_task_type(
    label=_(message='Launch all workflows for all documents'),
    dotted_path='mayan.apps.document_states.tasks.task_launch_all_workflows'
)
queue_workflows_slow.add_task_type(
    label=_(message='Launch a workflow'),
    dotted_path='mayan.apps.document_states.tasks.task_launch_workflow'
)



from rest_framework import generics as rest_framework_generics, status
from rest_framework.response import Response
from rest_framework.settings import api_settings

from django.core.exceptions import ImproperlyConfigured

from mayan.apps.dynamic_search.api_filters import RESTAPISearchFilter

from .api_view_mixins import (
    CheckQuerysetAPIViewMixin, DynamicFieldListAPIViewMixin,
    InstanceExtraDataAPIViewMixin, QuerySetOverrideCheckAPIViewMixin,
    SchemaInspectionAPIViewMixin, SerializerExtraContextAPIViewMixin
)
from .filters import MayanObjectPermissionsFilter, MayanSortingFilter
from .permissions import MayanPermission
from .serializers import BlankSerializer


class GenericAPIView(
    CheckQuerysetAPIViewMixin, SchemaInspectionAPIViewMixin,
    QuerySetOverrideCheckAPIViewMixin, rest_framework_generics.GenericAPIView
):
    filter_backends = (MayanObjectPermissionsFilter,)
    permission_classes = (MayanPermission,)
    request_method_real = None

    def initial(self, *args, **kwargs):
        # DRF modified the value of the request.method attribute.
        # Preserve the real request method for individual subclass usage.
        self.request_method_real = self.request.method.upper()
        result = super().initial(*args, **kwargs)
        return result


class CreateAPIView(
    CheckQuerysetAPIViewMixin, InstanceExtraDataAPIViewMixin,
    SchemaInspectionAPIViewMixin, SerializerExtraContextAPIViewMixin,
    QuerySetOverrideCheckAPIViewMixin, rest_framework_generics.CreateAPIView
):
    """
    requires:
        view_permission = {'POST': ...}
    """
    permission_classes = (MayanPermission,)


class ListAPIView(
    CheckQuerysetAPIViewMixin, DynamicFieldListAPIViewMixin,
    SerializerExtraContextAPIViewMixin, SchemaInspectionAPIViewMixin,
    QuerySetOverrideCheckAPIViewMixin, rest_framework_generics.ListAPIView
):
    """
    requires:
        object_permission = {'GET': ...}
    """
    filter_backends = (
        MayanObjectPermissionsFilter, MayanSortingFilter, RESTAPISearchFilter
    )
    # `permission_classes` is required for the `EventListAPIView`
    # when `Actions` objects support ACLs then this can be removed
    # as was intended.
    permission_classes = (MayanPermission,)


class ListCreateAPIView(
    CheckQuerysetAPIViewMixin, DynamicFieldListAPIViewMixin,
    InstanceExtraDataAPIViewMixin, SerializerExtraContextAPIViewMixin,
    SchemaInspectionAPIViewMixin, QuerySetOverrideCheckAPIViewMixin,
    rest_framework_generics.ListCreateAPIView
):
    """
    requires:
        object_permission = {'GET': ...}
        view_permission = {'POST': ...}
    """
    filter_backends = (
        MayanObjectPermissionsFilter, MayanSortingFilter, RESTAPISearchFilter
    )
    permission_classes = (MayanPermission,)


class ObjectActionAPIView(
    SerializerExtraContextAPIViewMixin, GenericAPIView
):
    action_response_status = None
    serializer_class = BlankSerializer

    def get_serializer_context(self):
        context = super().get_serializer_context()

        # When rendering the exception handler, DRF calls this class
        # method via `.override_method()` hiding the real request method.
        # Use the real request method instead which saved during the view
        # initialization for correct behavior.
        if self.request_method_real == 'POST':
            context.update(
                {
                    'object': self.get_object()
                }
            )
        return context

    def get_success_headers(self, data):
        try:
            return {
                'Location': str(
                    data[api_settings.URL_FIELD_NAME]
                )
            }
        except (TypeError, KeyError):
            return {}

    def object_action(self, serializer):
        raise ImproperlyConfigured(
            '{cls} class needs to specify the `.perform_action()` '
            'method.'.format(
                cls=self.__class__.__name__
            )
        )

    def post(self, request, *args, **kwargs):
        return self.view_action(request=request, *args, **kwargs)

    def view_action(self, request, *args, **kwargs):
        obj = self.get_object()

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        if hasattr(self, 'get_instance_extra_data'):
            for key, value in self.get_instance_extra_data().items():
                setattr(obj, key, value)

        result = self.object_action(
            obj=obj, request=request, serializer=serializer
        )

        if result:
            # If object action returned `serializer.data`.
            headers = self.get_success_headers(data=result)
            return Response(
                headers=headers, data=result,
                status=self.action_response_status or status.HTTP_200_OK
            )
        else:
            return Response(
                status=self.action_response_status or status.HTTP_200_OK
            )


class RetrieveAPIView(
    CheckQuerysetAPIViewMixin, DynamicFieldListAPIViewMixin,
    InstanceExtraDataAPIViewMixin, SerializerExtraContextAPIViewMixin,
    SchemaInspectionAPIViewMixin, QuerySetOverrideCheckAPIViewMixin,
    rest_framework_generics.RetrieveAPIView
):
    """
    requires:
        object_permission = {
            'GET': ...,
        }
    """
    filter_backends = (MayanObjectPermissionsFilter,)


class RetrieveDestroyAPIView(
    CheckQuerysetAPIViewMixin, DynamicFieldListAPIViewMixin,
    InstanceExtraDataAPIViewMixin, SerializerExtraContextAPIViewMixin,
    SchemaInspectionAPIViewMixin, QuerySetOverrideCheckAPIViewMixin,
    rest_framework_generics.RetrieveDestroyAPIView
):
    """
    requires:
        object_permission = {
            'DELETE': ...,
            'GET': ...,
        }
    """
    filter_backends = (MayanObjectPermissionsFilter,)


class RetrieveUpdateAPIView(
    CheckQuerysetAPIViewMixin, DynamicFieldListAPIViewMixin,
    InstanceExtraDataAPIViewMixin, SerializerExtraContextAPIViewMixin,
    SchemaInspectionAPIViewMixin, QuerySetOverrideCheckAPIViewMixin,
    rest_framework_generics.RetrieveUpdateAPIView
):
    """
    requires:
        object_permission = {
            'GET': ...,
            'PATCH': ...,
            'PUT': ...
        }
    """
    filter_backends = (MayanObjectPermissionsFilter,)


class RetrieveUpdateDestroyAPIView(
    CheckQuerysetAPIViewMixin, DynamicFieldListAPIViewMixin,
    InstanceExtraDataAPIViewMixin, SerializerExtraContextAPIViewMixin,
    SchemaInspectionAPIViewMixin, QuerySetOverrideCheckAPIViewMixin,
    rest_framework_generics.RetrieveUpdateDestroyAPIView
):
    """
    requires:
        object_permission = {
            'DELETE': ...,
            'GET': ...,
            'PATCH': ...,
            'PUT': ...
        }
    """
    filter_backends = (MayanObjectPermissionsFilter,)



from django.utils.translation import gettext_lazy as _

from mayan.apps.django_gpg.models import Key


class SignatureBaseModelBusinessLogicMixin:
    def get_key_id(self):
        if self.public_key_fingerprint:
            return self.public_key_fingerprint[-16:]
        else:
            return self.key_id

    def get_signature_type_display(self):
        if self.is_detached:
            return _(message='Detached')
        else:
            return _(message='Embedded')

    @property
    def is_detached(self):
        return hasattr(self, 'signature_file')

    @property
    def is_embedded(self):
        return not hasattr(self, 'signature_file')

    @property
    def key(self):
        try:
            return Key.objects.get(
                fingerprint=self.public_key_fingerprint
            )
        except Key.DoesNotExist:
            return None

    @property
    def key_algorithm(self):
        key = self.key

        if key:
            return key.algorithm

    @property
    def key_creation_date(self):
        key = self.key

        if key:
            return key.creation_date

    @property
    def key_expiration_date(self):
        key = self.key

        if key:
            return key.expiration_date

    @property
    def key_length(self):
        key = self.key

        if key:
            return key.length

    @property
    def key_type(self):
        key = self.key

        if key:
            return key.get_key_type_display()

    @property
    def key_user_id(self):
        key = self.key

        if key:
            return key.user_id



import hashlib

from django.apps import apps
from django.core import serializers
from django.db.models import F, Max, Q
from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.models.document_models import Document

from ..literals import WORKFLOW_ACTION_ON_ENTRY, WORKFLOW_ACTION_ON_EXIT


class WorkflowStateBusinessLogicMixin:
    @property
    def entry_actions(self):
        return self.actions.filter(when=WORKFLOW_ACTION_ON_ENTRY)

    @property
    def exit_actions(self):
        return self.actions.filter(when=WORKFLOW_ACTION_ON_EXIT)

    def get_actions_display(self):
        field_list = [
            str(field) for field in self.actions.all()
        ]
        field_list.sort()

        return ', '.join(field_list)

    get_actions_display.short_description = _(message='Actions')

    def get_escalations_display(self):
        field_list = [
            str(field) for field in self.escalations.all()
        ]
        field_list.sort()

        return ', '.join(field_list)

    get_escalations_display.short_description = _(message='Escalations')

    def get_documents(self):
        WorkflowInstanceLogEntry = apps.get_model(
            app_label='document_states',
            model_name='WorkflowInstanceLogEntry'
        )

        latest_entries = WorkflowInstanceLogEntry.objects.annotate(
            max_datetime=Max(
                'workflow_instance__log_entries__datetime'
            )
        ).filter(
            datetime=F('max_datetime')
        )

        state_latest_entries = latest_entries.filter(
            transition__destination_state=self
        )

        return Document.valid.filter(
            Q(
                workflows__pk__in=state_latest_entries.values_list(
                    'workflow_instance', flat=True
                )
            ) | Q(
                workflows__log_entries__isnull=True,
                workflows__workflow__states=self,
                workflows__workflow__states__initial=True
            )
        ).distinct()

    def get_hash(self):
        result = hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        )

        for action in self.actions.all():
            result.update(
                action.get_hash().encode()
            )

        for escalation in self.escalations.all():
            result.update(
                escalation.get_hash().encode()
            )

        return result.hexdigest()



import warnings

from django.utils.translation import gettext_lazy as _

from mayan.apps.common.apps import MayanAppConfig

from .literals import MESSAGE_SQLITE_WARNING
from .patches import patch_Migration
from .utils import check_for_sqlite
from .warnings import DatabaseWarning


class DatabasesApp(MayanAppConfig):
    has_tests = True
    name = 'mayan.apps.databases'
    verbose_name = _(message='Databases')

    def ready(self):
        super().ready()

        if check_for_sqlite():
            warnings.warn(
                category=DatabaseWarning,
                message=str(MESSAGE_SQLITE_WARNING)
            )

        patch_Migration()



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Locales'), name='locales'
)

event_user_locale_profile_edited = namespace.add_event_type(
    label=_(message='User locale profile edited'), name='user_locale_profile_edited'
)



from mayan.apps.smart_settings.classes import SettingNamespaceMigration
from mayan.apps.smart_settings.settings import setting_cluster
from mayan.apps.smart_settings.utils import smart_yaml_load

from .literals import (
    DEFAULT_DOCUMENTS_STORAGE_BACKEND,
    DEFAULT_DOCUMENTS_STORAGE_BACKEND_ARGUMENTS
)


class DocumentsSettingMigration(SettingNamespaceMigration):
    """
    0001 to 0002: Backend arguments are no longer quoted but YAML valid
                  too. Changed in version 3.3.
    0002 to 0003: Setting DOCUMENTS_RECENT_ADDED_COUNT renamed to
                  DOCUMENTS_RECENTLY_CREATED_COUNT,
                  DOCUMENTS_RECENT_ADDED_COUNT renamed to
                  DOCUMENTS_RECENTLY_CREATED_COUNT. Changed in version 4.0.
    0003 to 0004: New settings for document file storage, file page image
                  cache and version page image cache added and made to take
                  their initial settings from existing
                  DOCUMENTS_CACHE_STORAGE_BACKEND,
                  DOCUMENTS_CACHE_STORAGE_BACKEND_ARGUMENTS,
                  DOCUMENTS_STORAGE_BACKEND, and
                  DOCUMENTS_STORAGE_BACKEND_ARGUMENTS settings.
    """
    def documents_cache_storage_backend_arguments_0001(self, value):
        return smart_yaml_load(value=value)

    def documents_storage_backend_arguments_0001(self, value):
        return smart_yaml_load(value=value)

    def documents_file_page_image_cache_storage_backend_0003(self, value):
        # Get the setting by its new global name.
        setting = setting_cluster.get_setting(
            global_name='DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND'
        )
        # Load the value from the setting's old global name.
        setting.do_value_cache(global_name='DOCUMENTS_CACHE_STORAGE_BACKEND')
        return setting.value

    def documents_file_page_image_cache_storage_backend_arguments_0003(
        self, value
    ):
        # Get the setting by its new global name.
        setting = setting_cluster.get_setting(
            global_name='DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS'
        )
        # Load the value from the setting's old global name.
        setting.do_value_cache(
            global_name='DOCUMENTS_CACHE_STORAGE_BACKEND_ARGUMENTS'
        )
        return setting.value

    def documents_file_storage_backend_0003(self, value):
        # Get the setting by its new global name.
        setting = setting_cluster.get_setting(
            global_name='DOCUMENTS_FILE_STORAGE_BACKEND'
        )
        # Load the value from the setting's old global name.
        setting.do_value_cache(
            global_name='DOCUMENTS_STORAGE_BACKEND',
            default_override=DEFAULT_DOCUMENTS_STORAGE_BACKEND
        )
        return setting.value

    def documents_file_storage_backend_arguments_0003(self, value):
        # Get the setting by its new global name.
        setting = setting_cluster.get_setting(
            global_name='DOCUMENTS_FILE_STORAGE_BACKEND_ARGUMENTS'
        )
        # Load the value from the setting's old global name.
        setting.do_value_cache(
            global_name='DOCUMENTS_STORAGE_BACKEND_ARGUMENTS',
            default_override=DEFAULT_DOCUMENTS_STORAGE_BACKEND_ARGUMENTS
        )
        return setting.value

    def documents_recently_accessed_count_0002(self, value):
        # Get the setting by its new global name.
        setting = setting_cluster.get_setting(
            global_name='DOCUMENTS_RECENTLY_ACCESSED_COUNT'
        )
        # Load the value from the setting's old global name.
        setting.do_value_cache(global_name='DOCUMENTS_RECENT_ACCESS_COUNT')
        return setting.value

    def documents_recently_created_count_0002(self, value):
        # Get the setting by its new global name.
        setting = setting_cluster.get_setting(
            global_name='DOCUMENTS_RECENTLY_CREATED_COUNT'
        )
        # Load the value from the setting's old global name.
        setting.do_value_cache(global_name='DOCUMENTS_RECENT_ADDED_COUNT')
        return setting.value

    def documents_version_page_image_cache_storage_backend_0003(self, value):
        # Get the setting by its new global name.
        setting = setting_cluster.get_setting(
            global_name='DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND'
        )
        # Load the value from the setting's old global name.
        setting.do_value_cache(global_name='DOCUMENTS_CACHE_STORAGE_BACKEND')
        return setting.value

    def documents_version_page_image_cache_storage_backend_arguments_0003(
        self, value
    ):
        # Get the setting by its new global name.
        setting = setting_cluster.get_setting(
            global_name='DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS'
        )
        # Load the value from the setting's old global name.
        setting.do_value_cache(
            global_name='DOCUMENTS_CACHE_STORAGE_BACKEND_ARGUMENTS'
        )
        return setting.value



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0029_auto_20230505_1716')
    ]

    operations = [
        migrations.AlterField(
            field=models.TextField(
                blank=True, help_text='JSON encoded data for the '
                'backend class.', verbose_name='Backend data'
            ), model_name='workflowstateaction', name='backend_data'
        ),
        migrations.AlterField(
            field=models.CharField(
                help_text='The dotted Python path to the backend class.',
                max_length=128, verbose_name='Backend path'
            ), model_name='workflowstateaction', name='backend_path'
        )
    ]



import logging
import sys
import uuid

from django.utils.translation import gettext_lazy as _

from mayan.apps.common.apps import MayanAppConfig

from .backends.base import LockingBackend
from .literals import COMMAND_NAME_LOCK_MANAGER_PURGE_LOCKS, TEST_LOCK_NAME
from .settings import setting_backend

logger = logging.getLogger(name=__name__)


class LockManagerApp(MayanAppConfig):
    has_tests = True
    name = 'mayan.apps.lock_manager'
    verbose_name = _(message='Lock manager')

    def ready(self):
        super().ready()

        if COMMAND_NAME_LOCK_MANAGER_PURGE_LOCKS not in sys.argv:
            logger.debug('Starting lock backend connectivity test')
            # Don't test for locks during the `task_manager_purge_locks`
            # command as there may be some stuck locks which will block
            # the command.
            lock_name = '{}-{}'.format(
                TEST_LOCK_NAME, uuid.uuid4()
            )
            lock_instance = LockingBackend.get_backend()
            try:
                lock = lock_instance.acquire_lock(
                    name=lock_name, timeout=1
                )
            except Exception as exception:
                raise RuntimeError(
                    'Error initializing the locking backend: {}; {}'.format(
                        setting_backend.value, exception
                    )
                ) from exception
            else:
                lock.release()



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('linking', '0004_auto_20150708_0320'),
    ]

    operations = [
        migrations.AlterField(
            model_name='smartlink',
            name='dynamic_label',
            field=models.CharField(
                help_text="Enter a template to render. Use Django's default "
                "templating language (https://docs.djangoproject.com/en/1.7/"
                "ref/templates/builtins/). The {{ document }} context "
                "variable is available.", max_length=96,
                verbose_name='Dynamic label', blank=True
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='smartlinkcondition',
            name='expression',
            field=models.TextField(
                help_text="Enter a template to render. Use Django's default "
                "templating language (https://docs.djangoproject.com/en/1.7/"
                "ref/templates/builtins/). The {{ document }} context "
                "variable is available.", verbose_name='Expression'
            ),
            preserve_default=True,
        ),
    ]



from unittest import mock

from mayan.apps.testing.tests.base import BaseTestCase

from ..events import event_credential_used

from .mixins import StoredCredentialTestMixin


class StoredCredentialModelTestCase(
    StoredCredentialTestMixin, BaseTestCase
):
    _test_stored_credential_backend_path = 'mayan.apps.credentials.tests.credential_backends.TestCredentialBackendWithPostProcessing'

    @mock.patch(
        target='mayan.apps.credentials.tests.credential_backends.TestCredentialBackendWithPostProcessing.post_processing'
    )
    def test_credential_post_processing(self, mocked_method):
        self._clear_events()

        self._test_stored_credential.get_backend_data()

        self.assertEqual(mocked_method.call_count, 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_stored_credential)
        self.assertEqual(events[0].target, self._test_stored_credential)
        self.assertEqual(events[0].verb, event_credential_used.id)



from django.contrib import messages
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.sources.classes import DocumentCreateWizardStep
from mayan.apps.sources.wizard_steps import DocumentCreateWizardStepDocumentType

from .forms import DocumentMetadataFormSet
from .models import MetadataType
from .permissions import permission_document_metadata_add
from .utils import decode_metadata_from_query_string, save_metadata_list


class DocumentCreateWizardStepMetadata(DocumentCreateWizardStep):
    form_class = DocumentMetadataFormSet
    label = _(message='Enter document metadata')
    name = 'metadata_entry'
    number = 1

    @staticmethod
    def get_metadata_queryset_for_document(document, user=None):
        queryset_documents = Document.valid.all()

        if user:
            queryset_documents = AccessControlList.objects.restrict_queryset(
                permission=permission_document_metadata_add,
                queryset=queryset_documents, user=user
            )

        try:
            queryset_documents.get(pk=document.pk)
        except Document.DoesNotExist:
            queryset_metadata_types = MetadataType.objects.none()
        else:
            queryset_metadata_types = MetadataType.objects.all()

            if user:
                queryset_metadata_types = AccessControlList.objects.restrict_queryset(
                    permission=permission_document_metadata_add,
                    queryset=queryset_metadata_types,
                    user=user
                )

        return queryset_metadata_types

    @staticmethod
    def get_metadata_queryset_for_document_type(document_type, user):
        queryset_document_types = DocumentType.objects.all()

        queryset_document_types_restricted = AccessControlList.objects.restrict_queryset(
            permission=permission_document_metadata_add,
            queryset=queryset_document_types, user=user
        )

        try:
            queryset_document_types_restricted.get(pk=document_type.pk)
        except DocumentType.DoesNotExist:
            queryset_metadata_types = MetadataType.objects.none()
        else:
            queryset_metadata_types = MetadataType.objects.filter(
                document_types__document_type=document_type.pk
            )

            queryset_metadata_types = AccessControlList.objects.restrict_queryset(
                permission=permission_document_metadata_add,
                queryset=queryset_metadata_types, user=user
            )

        return queryset_metadata_types

    @classmethod
    def condition(cls, wizard):
        """
        Skip step if document type has no associated metadata
        """
        cleaned_data = wizard.get_cleaned_data_for_step(
            step=DocumentCreateWizardStepDocumentType.name
        ) or {}

        document_type = cleaned_data.get('document_type')

        if document_type:
            return document_type.metadata.exists()

    @classmethod
    def get_form_initial(cls, wizard):
        initial = []

        step_data = wizard.get_cleaned_data_for_step(
            step=DocumentCreateWizardStepDocumentType.name
        )

        user = wizard.request.user

        if step_data:
            document_type = step_data['document_type']

            queryset_metadata_types = DocumentCreateWizardStepMetadata.get_metadata_queryset_for_document_type(
                document_type=document_type, user=user
            )

            for metadata_type in queryset_metadata_types:
                initial.append(
                    {
                        'document_type': document_type,
                        'metadata_type': metadata_type
                    }
                )

        return initial

    @classmethod
    def done(cls, wizard):
        result = {}
        cleaned_data = wizard.get_cleaned_data_for_step(step=cls.name)
        if cleaned_data:
            for index, metadata_entry in enumerate(iterable=cleaned_data):
                if metadata_entry.get('update'):
                    result[
                        'metadata{}_metadata_type_id'.format(index)
                    ] = metadata_entry['metadata_type_id']
                    result[
                        'metadata{}_value'.format(index)
                    ] = metadata_entry['value']

        return result

    @classmethod
    def step_post_upload_process(
        cls, document, query_string, source_id, user_id
    ):
        User = get_user_model()

        try:
            user = User.objects.get(pk=user_id)
        except User.DoesNotExist:
            user = None

        if user:
            queryset_metadata_types_restricted = DocumentCreateWizardStepMetadata.get_metadata_queryset_for_document(
                document=document, user=user
            )

            metadata_dict_list = decode_metadata_from_query_string(
                query_string=query_string
            )

            # Convert metadata_dict_list and used to filter the restricted
            # queryset.
            queryset_metadata_types_restricted = queryset_metadata_types_restricted.filter(
                document_types__document_type=document.document_type.pk
            )

            # Convert back the metadata type queryset into a
            # metadata_dict_list.

            metadata_dict_list_restricted = []

            metadata_type_id_list_restricted = queryset_metadata_types_restricted.values_list(
                'id', flat=True
            )

            for metadata_entry in metadata_dict_list:
                if int(metadata_entry['metadata_type_id']) in metadata_type_id_list_restricted:
                    metadata_dict_list_restricted.append(metadata_entry)

            save_metadata_list(
                create=True, document=document,
                metadata_list=metadata_dict_list_restricted, user=user
            )

    @classmethod
    def get_next_is_enabled(cls, wizard):
        step_data = wizard.get_cleaned_data_for_step(
            step=DocumentCreateWizardStepDocumentType.name
        )

        user = wizard.request.user

        if step_data:
            document_type = step_data['document_type']

            queryset_metadata_types = cls.get_metadata_queryset_for_document_type(
                document_type=document_type, user=user
            )

            queryset_metadata_types_required = MetadataType.objects.get_for_document_type(
                document_type=document_type, required=True
            )

            has_required_metadata = not queryset_metadata_types_required.exclude(
                pk__in=queryset_metadata_types.values('pk')
            ).exists()

            if not has_required_metadata:
                messages.error(
                    message=_(
                        'One of more metadata types that are required for '
                        'this document type are not available.'
                    ), request=wizard.request
                )
                wizard.render_goto_step(goto_step=0)

                return False

        return True


DocumentCreateWizardStep.register(step=DocumentCreateWizardStepMetadata)



from django.contrib import messages
from django.shortcuts import get_object_or_404
from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import ConfirmView, SingleObjectListView

from ..icons import (
    icon_notification_list, icon_notification_mark_read,
    icon_notification_mark_read_all
)
from ..links import link_event_type_subscription_list

from .view_mixins import NotificationViewMixin


class NotificationListView(NotificationViewMixin, SingleObjectListView):
    view_icon = icon_notification_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_notification_list,
            'no_results_main_link': link_event_type_subscription_list.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Subscribe to global or object events to receive '
                'notifications.'
            ),
            'no_results_title': _(message='There are no notifications'),
            'title': _(message='Notifications')
        }


class NotificationMarkRead(NotificationViewMixin, ConfirmView):
    post_action_redirect = reverse_lazy(
        viewname='events:user_notifications_list'
    )
    view_icon = icon_notification_mark_read

    def get_extra_context(self):
        return {
            'title': _(message='Mark the selected notification as read?')
        }

    def get_object(self):
        return get_object_or_404(
            klass=self.get_source_queryset(),
            pk=self.kwargs['notification_id']
        )

    def view_action(self, form=None):
        obj = self.get_object()
        obj.read = True
        obj.save()

        messages.success(
            message=_(message='Notification marked as read.'), request=self.request
        )


class NotificationMarkReadAll(NotificationViewMixin, ConfirmView):
    post_action_redirect = reverse_lazy(
        viewname='events:user_notifications_list'
    )
    view_icon = icon_notification_mark_read_all

    def get_extra_context(self):
        return {
            'title': _(message='Mark all notification as read?')
        }

    def view_action(self, form=None):
        self.get_source_queryset().update(read=True)

        messages.success(
            message=_(message='All notifications marked as read.'),
            request=self.request
        )



from rest_framework.generics import get_object_or_404

from mayan.apps.acls.models import AccessControlList

from .models import Source


class ParentObjectSourceAPIViewMixin:
    def get_parent_queryset_source(self):
        return Source.objects.all()

    def get_source(self, permission=None):
        queryset = self.get_parent_queryset_source()

        if not permission:
            mayan_external_object_permission_map = getattr(
                self, 'mayan_external_object_permission_map', {}
            )

            permission = mayan_external_object_permission_map.get(
                self.request.method, None
            )

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs['source_id']
        )



from mayan.apps.common.tests.mixins import PropertyHelperTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from .mixins.base_mixins import SourceMetadataTestmixin


class SourceBackendDocumentPropertyHelperTestCase(
    SourceMetadataTestmixin, PropertyHelperTestMixin, BaseTestCase
):
    def test_basic(self):
        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document.source_metadata_value_of,
                self._test_source_metadata.key
            ), self._test_source_metadata.value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_no_file(self):
        self._test_document.file_latest.delete()

        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document.source_metadata_value_of,
                self._test_source_metadata.key
            ), None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_no_source_metadata(self):
        self._test_source_metadata.delete()

        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document.source_metadata_value_of,
                self._test_source_metadata.key
            ), None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_wrong_source_metadata_name(self):
        self._test_source_metadata.delete()

        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document.source_metadata_value_of, 'invalid'
            ), None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class SourceBackendDocumentFilePropertyHelperTestCase(
    SourceMetadataTestmixin, PropertyHelperTestMixin, BaseTestCase
):
    def test_basic(self):
        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document_file.source_metadata_value_of,
                self._test_source_metadata.key
            ), self._test_source_metadata.value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_no_source_metadata(self):
        self._test_source_metadata.delete()

        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document_file.source_metadata_value_of,
                self._test_source_metadata.key
            ), None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_wrong_source_metadata_name(self):
        self._clear_events()

        self.assertEqual(
            getattr(
                self._test_document_file.source_metadata_value_of, 'invalid'
            ), None
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import gettext_lazy as _

from actstream.models import Action, any_stream

from mayan.apps.views.generics import SingleObjectListView
from mayan.apps.views.view_mixins import ExternalContentTypeObjectViewMixin

from ..icons import (
    icon_event_list, icon_object_event_list, icon_verb_event_list
)
from ..permissions import permission_events_view

from .view_mixins import VerbEventViewMixin


class EventListBaseView(SingleObjectListView):
    object_permission = permission_events_view
    view_icon = icon_event_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_event_list,
            'no_results_text': _(
                'Events track actions that have been performed on, to, '
                'or with objects.'
            ),
            'no_results_title': _(message='There are no events'),
            'title': _(message='Events')
        }


class EventListView(EventListBaseView):
    view_icon = icon_event_list

    def get_source_queryset(self):
        return Action.objects.all()


class ObjectEventListView(
    ExternalContentTypeObjectViewMixin, EventListBaseView
):
    view_icon = icon_object_event_list

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'no_results_title': _(message='There are no events for this object'),
                'object': self.external_object,
                'title': _(message='Events for: %s') % self.external_object
            }
        )
        return context

    def get_source_queryset(self):
        return any_stream(obj=self.external_object)


class VerbEventListView(VerbEventViewMixin, EventListBaseView):
    view_icon = icon_verb_event_list

    def get_extra_context(self):
        context = super().get_extra_context()
        context.update(
            {
                'no_results_title': _(message='There are no events of this type'),
                'title': _(
                    'Events of type: %s'
                ) % self.event_type
            }
        )
        return context

    def get_source_queryset(self):
        return Action.objects.filter(verb=self.event_type.id)



class SourceBackendActionInterfaceArgument:
    def __eq__(self, other):
        # Allow quick deduplication by using `not in list`.
        return self.name == other.name

    def __init__(self, help_text=None, hidden=False, required=True, **kwargs):
        try:
            self.default = kwargs.pop('default')
        except KeyError:
            self.has_default = False
        else:
            self.has_default = True

        self.help_text = help_text
        self.hidden = hidden
        self.required = required



from django import forms
from django.utils.translation import gettext_lazy as _

from ..classes import DocumentVersionModification
from ..fields import DocumentVersionField
from ..models.document_version_models import DocumentVersion


class DocumentVersionForm(forms.ModelForm):
    class Meta:
        fields = ('comment',)
        model = DocumentVersion


class DocumentVersionModificationBackendForm(forms.Form):
    backend = forms.ChoiceField(
        choices=(), help_text=_(
            'The backend that will be executed against the selected '
            'document version.'
        ), label=_(message='Backend')
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['backend'].choices = DocumentVersionModification.get_choices()


class DocumentVersionPreviewForm(forms.Form):
    document_version = DocumentVersionField()

    def __init__(self, *args, **kwargs):
        document_version = kwargs.pop('instance', None)
        transformation_instance_list = kwargs.pop(
            'transformation_instance_list', ()
        )
        super().__init__(*args, **kwargs)
        self.fields['document_version'].initial = document_version
        self.fields['document_version'].widget.attrs.update(
            {
                'transformation_instance_list': transformation_instance_list
            }
        )



from django.db import migrations
from django.template.defaultfilters import slugify


def code_assign_slugs(apps, schema_editor):
    Index = apps.get_model(app_label='document_indexing', model_name='Index')

    for index in Index.objects.using(alias=schema_editor.connection.alias).all():
        index.slug = slugify(value=index.label)
        index.save()


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0005_index_slug')
    ]

    operations = [
        migrations.RunPython(code=code_assign_slugs)
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_key_delete, icon_key_detail, icon_key_download, icon_key_setup,
    icon_key_upload, icon_keyserver_search, icon_private_key_list,
    icon_public_key_list
)
from .permissions import (
    permission_key_delete, permission_key_download, permission_key_receive,
    permission_key_view, permission_key_upload, permission_keyserver_query
)

link_key_delete = Link(
    args=('resolved_object.pk',), icon=icon_key_delete,
    permission=permission_key_delete, tags='dangerous', text=_(message='Delete'),
    view='django_gpg:key_delete'
)
link_key_detail = Link(
    args=('resolved_object.pk',), icon=icon_key_detail,
    permission=permission_key_view, text=_(message='Details'),
    view='django_gpg:key_detail'
)
link_key_download = Link(
    args=('resolved_object.pk',), icon=icon_key_download,
    permission=permission_key_download, text=_(message='Download'),
    view='django_gpg:key_download'
)
link_key_query = Link(
    icon=icon_keyserver_search,
    permission=permission_keyserver_query, text=_(message='Query keyservers'),
    view='django_gpg:key_query'
)
link_key_receive = Link(
    args='object.key_id', keep_query=True,
    permission=permission_key_receive, text=_(message='Import'),
    view='django_gpg:key_receive'
)
link_key_setup = Link(
    icon=icon_key_setup, permission=permission_key_view,
    text=_(message='Key management'), view='django_gpg:key_public_list'
)
link_key_upload = Link(
    icon=icon_key_upload, permission=permission_key_upload,
    text=_(message='Upload key'), view='django_gpg:key_upload'
)
link_private_key_list = Link(
    icon=icon_private_key_list, permission=permission_key_view,
    text=_(message='Private keys'), view='django_gpg:key_private_list'
)
link_public_key_list = Link(
    icon=icon_public_key_list, permission=permission_key_view,
    text=_(message='Public keys'), view='django_gpg:key_public_list'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import (
    menu_list_facet, menu_object, menu_return, menu_secondary, menu_setup
)
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.views.column_widgets import TwoStateWidget

from .classes import Setting, SettingCluster, SettingNamespace
from .column_widgets import WidgetSettingValue
from .links import (
    link_setting_cluster_configuration_save,
    link_setting_cluster_namespace_list, link_setting_edit,
    link_setting_namespace_detail, link_setting_namespace_root_list,
    link_setting_revert
)
from .settings import setting_cluster
from .widgets import setting_widget


class SmartSettingsApp(MayanAppConfig):
    app_namespace = 'settings'
    app_url = 'settings'
    has_tests = True
    name = 'mayan.apps.smart_settings'
    verbose_name = _(message='Smart settings')

    def ready(self):
        super().ready()

        SettingCluster.load_modules()

        SourceColumn(
            func=lambda context: len(
                context['object'].setting_dict
            ), label=_(message='Setting count'), include_label=True,
            source=SettingNamespace
        )
        SourceColumn(
            func=lambda context: setting_widget(
                context['object']
            ), label=_(message='Name'), is_identifier=True, source=Setting
        )
        SourceColumn(
            attribute='get_value_current', include_label=True,
            label=_(message='Value'), widget=WidgetSettingValue,
            source=Setting
        )
        SourceColumn(
            attribute='get_is_overridden', include_label=True, source=Setting,
            widget=TwoStateWidget
        )
        SourceColumn(
            attribute='get_has_value_new', include_label=True,
            source=Setting, widget=TwoStateWidget
        )

        menu_list_facet.bind_links(
            links=(link_setting_namespace_detail,),
            sources=(SettingNamespace,)
        )
        menu_object.bind_links(
            links=(
                link_setting_edit, link_setting_revert
            ), sources=(Setting,)
        )
        menu_return.bind_links(
            links=(link_setting_namespace_root_list,), sources=(
                SettingNamespace, Setting,
                'settings:setting_cluster_namespace_list'
            )
        )
        menu_secondary.bind_links(
            links=(link_setting_cluster_configuration_save,), sources=(
                SettingNamespace, Setting,
                'settings:setting_cluster_namespace_list'
            )
        )
        menu_setup.bind_links(
            links=(link_setting_cluster_namespace_list,)
        )

        setting_cluster.do_settings_updated_clear()
        setting_cluster.do_post_edit_function_call()
        setting_cluster.do_last_known_good_save()



from mayan.apps.document_states.permissions import permission_workflow_template_edit
from mayan.apps.document_states.tests.mixins.workflow_template_state_action_mixins import (
    WorkflowTemplateStateActionTestMixin,
    WorkflowTemplateStateActionViewTestMixin
)
from mayan.apps.testing.tests.base import BaseTestCase, GenericViewTestCase

from ..models import LayerTransformation
from ..workflow_actions import TransformationAddAction

from .literals import (
    TEST_TRANSFORMATION_ROTATE_ARGUMENT, TEST_TRANSFORMATION_ROTATE_NAME
)


class TransformationActionTestCase(
    WorkflowTemplateStateActionTestMixin, BaseTestCase
):
    def setUp(self):
        super().setUp()

        self._test_document.delete()
        self._test_document.delete()

        self._upload_test_document()

    def test_transformation_add_pages_all_action(self):
        transformation_count = LayerTransformation.objects.get_for_object(
            obj=self._test_document.pages.first()
        ).count()

        self._execute_workflow_template_state_action(
            klass=TransformationAddAction, kwargs={
                'pages': '',
                'transformation_class': TEST_TRANSFORMATION_ROTATE_NAME,
                'transformation_arguments': TEST_TRANSFORMATION_ROTATE_ARGUMENT
            }
        )
        self.assertEqual(
            LayerTransformation.objects.get_for_object(
                obj=self._test_document.pages.first()
            ).count(), transformation_count + 1
        )

    def test_transformation_add_pages_first_action(self):
        transformation_count = LayerTransformation.objects.get_for_object(
            obj=self._test_document.pages.first()
        ).count()

        self._execute_workflow_template_state_action(
            klass=TransformationAddAction, kwargs={
                'pages': '1',
                'transformation_class': TEST_TRANSFORMATION_ROTATE_NAME,
                'transformation_arguments': TEST_TRANSFORMATION_ROTATE_ARGUMENT
            }
        )
        self.assertEqual(
            LayerTransformation.objects.get_for_object(
                obj=self._test_document.pages.first()
            ).count(), transformation_count + 1
        )


class TransformationActionViewTestCase(
    WorkflowTemplateStateActionViewTestMixin, GenericViewTestCase
):
    def test_transformation_add_pages_all_action_create_view(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        response = self._request_test_workflow_template_state_action_create_post_view(
            backend_path='mayan.apps.converter.workflow_actions.TransformationAddAction',
            extra_data={
                'pages': '',
                'transformation_class': TEST_TRANSFORMATION_ROTATE_NAME,
                'transformation_arguments': TEST_TRANSFORMATION_ROTATE_ARGUMENT
            }
        )
        self.assertEqual(response.status_code, 302)

    def test_transformation_add_pages_first_action_create_view(self):
        self.grant_access(
            obj=self._test_workflow_template,
            permission=permission_workflow_template_edit
        )

        response = self._request_test_workflow_template_state_action_create_post_view(
            backend_path='mayan.apps.converter.workflow_actions.TransformationAddAction',
            extra_data={
                'pages': '1',
                'transformation_class': TEST_TRANSFORMATION_ROTATE_NAME,
                'transformation_arguments': TEST_TRANSFORMATION_ROTATE_ARGUMENT
            }
        )
        self.assertEqual(response.status_code, 302)



from django.db import migrations, models
import django.db.models.deletion


class AddFieldDynamicDefault(migrations.AddField):
    """
    Subclass of migrations.AddField to allow passing a calculated default
    value from another model instance.
    """
    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        StoredDuplicateBackend = from_state.apps.get_model(
            app_label='duplicates', model_name='StoredDuplicateBackend'
        )

        stored_backend = StoredDuplicateBackend.objects.using(
            alias=schema_editor.connection.alias
        ).first()

        if stored_backend:
            stored_backend_id = stored_backend.pk

            self.field.default = stored_backend_id

        super().database_forwards(
            app_label=app_label, schema_editor=schema_editor,
            from_state=from_state, to_state=to_state
        )


class Migration(migrations.Migration):
    dependencies = [
        ('duplicates', '0005_auto_20201130_0747')
    ]

    operations = [
        AddFieldDynamicDefault(
            model_name='duplicateddocument',
            name='stored_backend',
            field=models.ForeignKey(
                default=None, on_delete=django.db.models.deletion.CASCADE,
                related_name='duplicate_entries',
                to='duplicates.StoredDuplicateBackend',
                verbose_name='Stored duplicate backend'
            ),
            preserve_default=False
        )
    ]



import logging
from pathlib import Path

from django.apps import apps
from django.contrib.auth import get_user_model
from django.core.files import File
from django.db import OperationalError

from mayan.apps.storage.tasks import task_shared_upload_delete
from mayan.celery import app

from ..literals import DEFAULT_DOCUMENT_FILE_ACTION_NAME

from .utils import execute_callback

logger = logging.getLogger(name=__name__)


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_document_file_checksum_update(
    self, document_file_id, action_name=None, callback_dict=None,
    is_document_upload_sequence=False, user_id=None
):
    callback_dict = callback_dict or {}

    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )

    try:
        document_file = DocumentFile.objects.get(pk=document_file_id)
    except OperationalError as exception:
        raise self.retry(exc=exception)

    try:
        document_file.checksum_update(save=False)
        document_file._event_ignore = True
        document_file.save(
            update_fields=('checksum',)
        )
    except OperationalError as exception:
        logger.warning(
            'Operational error during attempt to update the checksum for '
            'document file: %s; %s. Retrying.', document_file,
            exception
        )
        raise self.retry(exc=exception)
    except Exception as exception:
        logger.critical(
            'Unexpected exception when updating checksum for document '
            'file: %s; %s.', document_file, exception
        )
        raise
    else:
        if is_document_upload_sequence:
            task_document_file_mimetype_update.apply_async(
                kwargs={
                    'action_name': action_name,
                    'callback_dict': callback_dict,
                    'document_file_id': document_file.pk,
                    'is_document_upload_sequence': is_document_upload_sequence,
                    'user_id': user_id
                }
            )


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_document_file_create(
    self, document_id, shared_uploaded_file_id, action_name=None,
    callback_dict=None, comment=None, filename=None,
    is_document_upload_sequence=False, user_id=None
):
    callback_dict = callback_dict or {}

    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )
    SharedUploadedFile = apps.get_model(
        app_label='storage', model_name='SharedUploadedFile'
    )
    User = get_user_model()

    action_name = action_name or DEFAULT_DOCUMENT_FILE_ACTION_NAME

    try:
        document = Document.objects.get(pk=document_id)
        shared_uploaded_file = SharedUploadedFile.objects.get(
            pk=shared_uploaded_file_id
        )

        if user_id:
            user = User.objects.get(pk=user_id)
        else:
            user = None
    except OperationalError as exception:
        raise self.retry(exc=exception)

    comment = comment or ''
    filename = filename or Path(
        str(shared_uploaded_file)
    ).name

    with shared_uploaded_file.open() as file_object:
        try:
            document_file = DocumentFile(
                comment=comment, document=document,
                file=File(file=file_object), filename=filename
            )
            document_file._event_actor = user
            document_file.save(skip_introspection=True)
        except OperationalError as exception:
            logger.error(
                'Operational error while uploading new file for '
                'document: %s; %s. Retrying.', document, exception
            )
            raise self.retry(exc=exception)
        except Exception as exception:
            logger.critical(
                'Unexpected exception when uploading new file for '
                'document: %s; %s.', document, exception
            )
            raise
        else:
            task_shared_upload_delete.apply_async(
                kwargs={
                    'shared_uploaded_file_id': shared_uploaded_file.pk
                }
            )

            execute_callback(
                callback_dict=callback_dict, document_file=document_file,
                name='post_document_file_create'
            )

            task_document_file_size_update.apply_async(
                kwargs={
                    'action_name': action_name,
                    'callback_dict': callback_dict,
                    'document_file_id': document_file.pk,
                    'is_document_upload_sequence': is_document_upload_sequence,
                    'user_id': user_id
                }
            )


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_document_file_delete(self, document_file_id, user_id=None):
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )
    User = get_user_model()

    try:
        if user_id:
            user = User.objects.get(pk=user_id)
        else:
            user = None

        document_file = DocumentFile.objects.get(
            pk=document_file_id
        )
    except OperationalError as exception:
        raise self.retry(exc=exception)

    try:
        document_file._event_actor = user
        document_file._event_keep_attributes = ('_event_actor',)
        document_file.delete()
    except OperationalError as exception:
        raise self.retry(exc=exception)


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_document_file_mimetype_update(
    self, document_file_id, action_name=None, callback_dict=None,
    is_document_upload_sequence=False, user_id=None
):
    callback_dict = callback_dict or {}

    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )

    try:
        document_file = DocumentFile.objects.get(pk=document_file_id)
    except OperationalError as exception:
        raise self.retry(exc=exception)

    try:
        document_file.mimetype_update(save=False)
        document_file._event_ignore = True
        document_file.save(
            update_fields=('encoding', 'mimetype',)
        )
    except OperationalError as exception:
        logger.warning(
            'Operational error during attempt to update the MIME type of '
            'document file: %s; %s. Retrying.', document_file, exception
        )
        raise self.retry(exc=exception)
    except Exception as exception:
        logger.critical(
            'Unexpected exception when updating the MIME type of document '
            'file: %s; %s.', document_file, exception
        )
        raise
    else:
        if is_document_upload_sequence:
            task_document_file_page_count_update.apply_async(
                kwargs={
                    'action_name': action_name,
                    'callback_dict': callback_dict,
                    'document_file_id': document_file.pk,
                    'is_document_upload_sequence': is_document_upload_sequence,
                    'user_id': user_id
                }
            )


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_document_file_page_count_update(
    self, document_file_id, action_name=None, callback_dict=None,
    is_document_upload_sequence=False, user_id=None
):
    callback_dict = callback_dict or {}

    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )
    User = get_user_model()

    try:
        document_file = DocumentFile.objects.get(pk=document_file_id)

        if user_id:
            user = User.objects.get(pk=user_id)
        else:
            user = None
    except OperationalError as exception:
        raise self.retry(exc=exception)

    try:
        document_file.page_count_update(user=user)
    except OperationalError as exception:
        logger.warning(
            'Operational error during attempt to update the page count for '
            'document file: %s; %s. Retrying.', document_file,
            exception
        )
        raise self.retry(exc=exception)
    except Exception as exception:
        logger.critical(
            'Unexpected exception updating the page count of document '
            'file: %s; %s.', document_file, exception
        )
        raise
    else:
        document_file.upload_complete()

        if is_document_upload_sequence:
            task_document_file_version_create.apply_async(
                kwargs={
                    'action_name': action_name,
                    'document_file_id': document_file.pk,
                    'user_id': user_id
                }
            )

            execute_callback(
                callback_dict=callback_dict, document_file=document_file,
                name='post_document_file_upload'
            )


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_document_file_size_update(
    self, document_file_id, action_name=None, callback_dict=None,
    is_document_upload_sequence=False, user_id=None
):
    callback_dict = callback_dict or {}

    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )

    try:
        document_file = DocumentFile.objects.get(pk=document_file_id)
    except OperationalError as exception:
        raise self.retry(exc=exception)

    try:
        document_file.size_update(save=False)
        document_file._event_ignore = True
        document_file.save(
            update_fields=('size',)
        )
    except OperationalError as exception:
        logger.warning(
            'Operational error during attempt to update the size for'
            'document file: %s; %s. Retrying.', document_file,
            exception
        )
        raise self.retry(exc=exception)
    except Exception as exception:
        logger.critical(
            'Unexpected exception updating the size document '
            'file: %s; %s.', document_file, exception
        )
        raise
    else:
        if is_document_upload_sequence:
            task_document_file_checksum_update.apply_async(
                kwargs={
                    'action_name': action_name,
                    'callback_dict': callback_dict,
                    'document_file_id': document_file.pk,
                    'is_document_upload_sequence': is_document_upload_sequence,
                    'user_id': user_id
                }
            )


@app.task(ignore_result=True)
def task_document_file_upload(
    document_id, shared_uploaded_file_id, action_name=None,
    callback_dict=None, comment=None, expand=False, filename=None,
    user_id=None
):
    callback_dict = callback_dict or {}

    task_document_file_create.apply_async(
        kwargs={
            'action_name': action_name,
            'callback_dict': callback_dict,
            'comment': comment,
            'document_id': document_id,
            'filename': filename,
            'is_document_upload_sequence': True,
            'shared_uploaded_file_id': shared_uploaded_file_id,
            'user_id': user_id
        }
    )


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_document_file_version_create(
    self, action_name, document_file_id, comment=None, user_id=None
):
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )
    User = get_user_model()

    try:
        document_file = DocumentFile.objects.get(
            pk=document_file_id
        )

        if user_id:
            user = User.objects.get(pk=user_id)
        else:
            user = None
    except OperationalError as exception:
        raise self.retry(exc=exception)

    comment = comment or ''

    try:
        document_file.versions_new(
            action_name=action_name, comment=comment, user=user
        )
    except OperationalError as exception:
        logger.warning(
            'Operational error during attempt execute the document version '
            'action: %s for %s; %s. Retrying.', action_name, document_file,
            exception
        )
        raise self.retry(exc=exception)



from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.serializers.document_serializers import (
    DocumentSerializer
)
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from .models import Tag
from .permissions import (
    permission_tag_attach, permission_tag_create, permission_tag_delete,
    permission_tag_edit, permission_tag_remove, permission_tag_view
)
from .serializers import (
    DocumentTagAttachSerializer, DocumentTagRemoveSerializer, TagSerializer
)


class APITagDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected tag.
    get: Return the details of the selected tag.
    patch: Edit the selected tag.
    put: Edit the selected tag.
    """
    lookup_url_kwarg = 'tag_id'
    mayan_object_permission_map = {
        'DELETE': permission_tag_delete,
        'GET': permission_tag_view,
        'PATCH': permission_tag_edit,
        'PUT': permission_tag_edit
    }
    serializer_class = TagSerializer
    source_queryset = Tag.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APITagListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the tags.
    post: Create a new tag.
    """
    mayan_object_permission_map = {'GET': permission_tag_view}
    mayan_view_permission_map = {'POST': permission_tag_create}
    ordering_fields = ('id', 'label')
    serializer_class = TagSerializer
    source_queryset = Tag.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APITagDocumentListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the documents tagged by a particular tag.
    """
    external_object_class = Tag
    external_object_pk_url_kwarg = 'tag_id'
    mayan_external_object_permission_map = {'GET': permission_tag_view}
    mayan_object_permission_map = {'GET': permission_document_view}
    serializer_class = DocumentSerializer

    def get_source_queryset(self):
        return Document.valid.filter(
            pk__in=self.get_external_object().documents.only('pk')
        )


class APIDocumentTagAttachView(generics.ObjectActionAPIView):
    """
    post: Attach a tag to a document.
    """
    lookup_url_kwarg = 'document_id'
    mayan_object_permission_map = {'POST': permission_tag_attach}
    serializer_class = DocumentTagAttachSerializer
    source_queryset = Document.valid.all()

    def object_action(self, obj, request, serializer):
        tag = serializer.validated_data['tag']
        tag.attach_to(document=obj, user=self.request.user)


class APIDocumentTagRemoveView(generics.ObjectActionAPIView):
    """
    post: Remove a tag from a document.
    """
    lookup_url_kwarg = 'document_id'
    mayan_object_permission_map = {'POST': permission_tag_remove}
    serializer_class = DocumentTagRemoveSerializer
    source_queryset = Document.valid.all()

    def object_action(self, obj, request, serializer):
        tag = serializer.validated_data['tag']
        tag.remove_from(document=obj, user=self.request.user)


class APIDocumentTagListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the tags attached to a document.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    mayan_external_object_permission_map = {'GET': permission_tag_view}
    mayan_object_permission_map = {'GET': permission_tag_view}
    serializer_class = TagSerializer

    def get_source_queryset(self):
        return self.get_external_object().tags.all()



from mayan.apps.source_interactive.source_backend_actions.callback_mixins import (
    SourceBackendActionMixinCallbackDocumentUploadInteractive
)
from mayan.apps.source_interactive.source_backend_actions.interactive_actions import (
    SourceBackendActionInteractiveBase
)
from mayan.apps.sources.source_backend_actions.mixins.document_mixins import (
    SourceBackendActionMixinDocumentUploadInteractive
)
from mayan.apps.sources.source_backend_actions.mixins.document_type_mixins import (
    SourceBackendActionMixinDocumentTypeInteractive
)

from ..source_backend_actions.mixins import SourceBackendActionMixinCompressedInteractive


class SourceBackendActionDocumentUploadBasicInteractiveCompressed(
    SourceBackendActionMixinDocumentUploadInteractive,
    SourceBackendActionMixinCompressedInteractive,
    SourceBackendActionMixinCallbackDocumentUploadInteractive,
    SourceBackendActionMixinDocumentTypeInteractive,
    SourceBackendActionInteractiveBase
):
    """
    Minimal action for the test source.
    """
    name = 'document_upload'



from django import forms

from .widgets import (
    DocumentFilePagesCarouselWidget, DocumentVersionPagesCarouselWidget,
    PageImageWidget, ThumbnailFormWidget
)


class DocumentFileField(forms.fields.Field):
    widget = DocumentFilePagesCarouselWidget


class DocumentFilePageField(forms.fields.Field):
    widget = PageImageWidget


class DocumentVersionField(forms.fields.Field):
    widget = DocumentVersionPagesCarouselWidget


class DocumentVersionPageField(forms.fields.Field):
    widget = PageImageWidget


class ThumbnailFormField(forms.fields.Field):
    widget = ThumbnailFormWidget



from django.core.exceptions import ImproperlyConfigured
from django.core.files import File
from django.utils.translation import gettext_lazy as _
from django.http import StreamingHttpResponse

from rest_framework import status
from rest_framework.response import Response

from mayan.apps.converter.classes import ConverterBase
from mayan.apps.converter.utils import IndexedDictionary
from mayan.apps.sources.source_backend_actions.interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestView,
    SourceBackendActionInterfaceRequestViewForm,
    SourceBackendActionInterfaceTask
)
from mayan.apps.storage.models import SharedUploadedFile

from .arguments import (
    argument_encoded_filename, argument_file_cleanup,
    argument_file_identifier, argument_maximum_layer_order,
    argument_transformation_instance_list, argument_user
)


class SourceBackendActionMixinFileStoredDeleteBase:
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                encoded_filename = argument_encoded_filename

            def process_interface_context(self):
                super().process_interface_context()
                self.action_kwargs['encoded_filename'] = self.context[
                    'encoded_filename'
                ]

    def _execute(self, encoded_filename, **kwargs):
        source_backend_instance = self.source.get_backend_instance()

        source_backend_instance.action_file_delete(
            encoded_filename=encoded_filename
        )


class SourceBackendActionMixinFileStoredDeleteInteractive(
    SourceBackendActionMixinFileStoredDeleteBase
):
    class Interface:
        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            class Argument:
                encoded_filename = argument_encoded_filename

            def process_interface_context(self):
                super().process_interface_context()
                self.action_kwargs['encoded_filename'] = self.context[
                    'encoded_filename'
                ]

        class View(SourceBackendActionInterfaceRequestView):
            class Argument:
                encoded_filename = argument_encoded_filename

            def get_confirmation_context(self, encoded_filename, **kwargs):
                source_backend_instance = self.action.source.get_backend_instance()

                stored_file = source_backend_instance.get_stored_file(
                    encoded_filename=encoded_filename
                )

                context = {
                    'delete_view': True,
                    'object': stored_file,
                    'object_name': _(message='Stored file'),
                    'title': _(message='Delete stored file "%s"?') % stored_file,
                }

                return context

            def process_interface_context(self):
                super().process_interface_context()
                self.action_kwargs['encoded_filename'] = self.context[
                    'encoded_filename'
                ]

    def _execute(self, encoded_filename, **kwargs):
        source_backend_instance = self.source.get_backend_instance()

        source_backend_instance.action_file_delete(
            encoded_filename=encoded_filename
        )


class SourceBackendActionMixinFileStoredImage:
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                encoded_filename = argument_encoded_filename
                maximum_layer_order = argument_maximum_layer_order
                transformation_instance_list = argument_transformation_instance_list
                user = argument_user

            def process_action_data(self):
                super().process_action_data()

                self.interface_result = self.action_data()

            def process_interface_context(self):
                super().process_interface_context()
                self.action_kwargs['encoded_filename'] = self.context['encoded_filename']
                self.action_kwargs['maximum_layer_order'] = self.context['maximum_layer_order']
                self.action_kwargs['transformation_instance_list'] = self.context['transformation_instance_list']
                self.action_kwargs['user'] = self.context['user']

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            class Argument:
                encoded_filename = argument_encoded_filename

            def process_action_data(self):
                super().process_action_data()

                content_type = ConverterBase.get_output_content_type()

                self.interface_result = StreamingHttpResponse(
                    content_type=content_type,
                    streaming_content=self.action_data()
                )

            def process_interface_context(self):
                super().process_interface_context()
                # The request returns a list even though this is a single
                # value variable.
                self.action_kwargs['encoded_filename'] = self.context['encoded_filename'][0]

                query_dict = self.context['request'].GET

                self.action_kwargs['transformation_instance_list'] = IndexedDictionary(
                    dictionary=query_dict
                ).as_instance_list()

                self.action_kwargs['maximum_layer_order'] = query_dict.get(
                    'maximum_layer_order'
                )
                self.action_kwargs['user'] = self.context['request'].user

    def _execute(
        self, encoded_filename, maximum_layer_order=None,
        transformation_instance_list=None, user=None
    ):
        source_backend_instance = self.source.get_backend_instance()

        return source_backend_instance.action_file_image(
            encoded_filename=encoded_filename,
            maximum_layer_order=maximum_layer_order,
            transformation_instance_list=transformation_instance_list,
            user=user
        )


class SourceBackendActionMixinFileStoredList:
    stored_file_list_method_name = 'action_file_list'

    class Interface:
        class Model(SourceBackendActionInterface):
            def process_action_data(self):
                super().process_action_data()

                self.interface_result = self.action_data

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            """
            Empty interface to enable it for the action.
            """
            def process_action_data(self):
                # Typecast generator to list to allow List API test to pass.
                self.action_data = list(self.action_data)
                super().process_action_data()

    def _execute(self):
        source_backend_instance = self.source.get_backend_instance()

        source_stored_file_get_method = getattr(
            source_backend_instance, self.stored_file_list_method_name
        )

        return source_stored_file_get_method()


class SourceBackendActionMixinFileStoredBase:
    default_file_cleanup = False
    stored_file_identifier_name = 'file_identifier'
    stored_method_name_file_cleanup = 'action_file_cleanup'
    stored_method_name_file_get = 'action_file_get'

    class Interface:
        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                file_cleanup = argument_file_cleanup

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['file_cleanup'] = self.context['file_cleanup']

    # Default `file_cleanup` needed by email sources.
    def _background_task(self, file_identifier, file_cleanup=None, **kwargs):
        result = super()._background_task(**kwargs)
        # Source returned a None `file_identifier` which means that the
        # source file list is empty to short circuit a quick exit.
        if file_identifier is None:
            result['server_upload_entry_list'] = ()
            return result

        source_backend_instance = self.source.get_backend_instance()

        kwargs = {
            self.stored_file_identifier_name: file_identifier
        }

        source_stored_file_get_method = getattr(
            source_backend_instance, self.stored_method_name_file_get
        )

        result['server_upload_entry_list'] = []

        server_upload_entry_generator = source_stored_file_get_method(**kwargs)

        if server_upload_entry_generator is None:
            raise ImproperlyConfigured(
                'Source backend method `{}` must return an iterator of at '
                'least one element.'.format(
                    self.stored_method_name_file_get
                )
            )

        while True:
            try:
                server_upload_entry = next(server_upload_entry_generator)

                result['server_upload_entry_list'].append(
                    self.process_server_upload_entry(
                        server_upload_entry=server_upload_entry
                    )
                )
            except StopIteration:
                """
                No more files to process.
                """
                break

        if file_cleanup is None:
            file_cleanup = self.default_file_cleanup

        if file_cleanup:
            source_stored_file_cleanup_method = getattr(
                source_backend_instance, self.stored_method_name_file_cleanup,
                None
            )

            if source_stored_file_cleanup_method:
                source_stored_file_cleanup_method(**kwargs)

        return result

    # Default `file_cleanup` needed by email sources.
    def get_task_kwargs(self, file_identifier, file_cleanup=None, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        result['action_interface_kwargs'].update(
            {
                'file_cleanup': file_cleanup,
                'file_identifier': file_identifier
            }
        )

        return result

    def process_server_upload_entry(self, server_upload_entry):
        file_object = server_upload_entry.pop('file')

        with file_object.open(mode='rb') as file_object:
            shared_uploaded_file = SharedUploadedFile.objects.create(
                file=File(
                    file=file_object
                )
            )

            server_upload_entry['shared_uploaded_file_id'] = shared_uploaded_file.pk
            return server_upload_entry


class SourceBackendActionMixinFileStoredInteractive(
    SourceBackendActionMixinFileStoredBase
):
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                file_identifier = argument_file_identifier

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['file_identifier'] = self.context[
                    'file_identifier'
                ]

                source_backend_instance = self.action.source.get_backend_instance()

                self.action_kwargs['file_cleanup'] = source_backend_instance.kwargs.get(
                    'delete_after_upload', False
                )

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            def process_action_data(self):
                super().process_action_data()

                self.interface_result = Response(
                    status=status.HTTP_202_ACCEPTED
                )

            def process_interface_context(self):
                super().process_interface_context()
                self.action_kwargs['file_identifier'] = self.context[
                    self.action.stored_file_identifier_name
                ]

                source_backend_instance = self.action.source.get_backend_instance()

                self.action_kwargs['file_cleanup'] = source_backend_instance.kwargs.get(
                    'delete_after_upload', False
                )

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                file_identifier = argument_file_identifier

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['file_identifier'] = self.context[
                    'file_identifier'
                ]

                source_backend_instance = self.action.source.get_backend_instance()

                self.action_kwargs['file_cleanup'] = source_backend_instance.kwargs.get(
                    'delete_after_upload', False
                )

        class View(SourceBackendActionInterfaceRequestViewForm):
            def process_interface_context(self):
                super().process_interface_context()

                form = self.context['forms']['source_form']

                stored_source_file_id = form.cleaned_data['stored_source_file_id']

                self.action_kwargs['file_identifier'] = stored_source_file_id

                source_backend_instance = self.action.source.get_backend_instance()

                self.action_kwargs['file_cleanup'] = source_backend_instance.kwargs.get(
                    'delete_after_upload', False
                )


class SourceBackendActionMixinFileStoredInteractiveNot(
    SourceBackendActionMixinFileStoredBase
):
    class Interface:
        class Model(SourceBackendActionInterface):
            def process_interface_context(self):
                super().process_interface_context()

                source_backend_instance = self.action.source.get_backend_instance()

                self.action_kwargs['file_identifier'] = source_backend_instance.get_file_identifier()

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                file_cleanup = argument_file_cleanup
                file_identifier = argument_file_identifier

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['file_cleanup'] = self.context['file_cleanup']
                self.action_kwargs['file_identifier'] = self.context['file_identifier']

    def convert_dry_run_to_file_cleanup(self, dry_run):
        if dry_run is True:
            return False
        elif dry_run is False:
            return True
        elif dry_run is None:
            return True

    def get_task_kwargs(self, dry_run, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        result['action_interface_kwargs']['file_cleanup'] = self.convert_dry_run_to_file_cleanup(
            dry_run=dry_run
        )

        return result



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_error_log_deleted
from ..permissions import (
    permission_error_log_entry_delete, permission_error_log_entry_view
)

from .mixins import (
    ErrorLogPartitionEntryAPIViewTestMixin, ErrorLogPartitionEntryTestMixin
)


class ErrorLogPartitionEntryAPIViewTestCase(
    ErrorLogPartitionEntryTestMixin, ErrorLogPartitionEntryAPIViewTestMixin,
    BaseAPITestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_error_log_object()
        self._create_test_error_log_entry()

    def test_error_log_delete_api_view_no_permission(self):
        error_log_entry_count = self._test_object.error_log.count()

        self._clear_events()

        response = self._request_test_error_log_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_object.error_log.count(), error_log_entry_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_error_log_delete_api_view_with_access(self):
        self.grant_access(
            self._test_object, permission=permission_error_log_entry_delete
        )

        error_log_entry_count = self._test_object.error_log.count()

        self._clear_events()

        response = self._request_test_error_log_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(
            self._test_object.error_log.count(), error_log_entry_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_object)
        self.assertEqual(events[0].verb, event_error_log_deleted.id)

    def test_error_log_detail_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_error_log_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_error_log_detail_api_view_with_access(self):
        self.grant_access(
            obj=self._test_object, permission=permission_error_log_entry_view
        )

        self._clear_events()

        response = self._request_test_error_log_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_error_log_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_error_log_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_error_log_list_api_view_with_access(self):
        self.grant_access(
            obj=self._test_object, permission=permission_error_log_entry_view
        )

        self._clear_events()

        response = self._request_test_error_log_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



TEST_SOURCE_SANE_SCANNER_FILE_CHECKSUM = 'e526d9deec60afeb84e6aa357d7815f4ecff6918e182d2398cb37b22e87c2bd0'

TEST_SOURCE_BACKEND_PATH_SANE_SCANNER = 'mayan.apps.source_sane_scanners.source_backends.SourceBackendSANEScanner'



import re

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('credentials', '0002_auto_20201215_0523')
    ]

    operations = [
        migrations.AddField(
            model_name='storedcredential', name='internal_name',
            field=models.CharField(
                blank=True, db_index=True, help_text='This value will be '
                'used by other apps to reference this credential. Can '
                'only contain letters, numbers, and underscores.',
                max_length=255, null=True, unique=True, validators=[
                    django.core.validators.RegexValidator(
                        re.compile('^[a-zA-Z0-9_]+\\Z'),
                        "Enter a valid 'internal name' consisting of "
                        "letters, numbers, and underscores.", 'invalid'
                    )
                ], verbose_name='Internal name')
        ),
        migrations.AlterField(
            model_name='storedcredential', name='backend_data',
            field=models.TextField(
                blank=True, help_text='JSON encoded data for the '
                'backend class.', verbose_name='Backend data'
            )
        )
    ]



import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('file_caching', '0009_alter_cache_options')
    ]

    operations = [
        migrations.AlterField(
            model_name='cache', name='maximum_size',
            field=models.PositiveBigIntegerField(
                help_text='Maximum size of the cache in bytes.',
                validators=[
                    django.core.validators.MinValueValidator(
                        limit_value=1
                    )
                ], verbose_name='Maximum size'
            )
        )
    ]



import importlib
import logging

from mayan.apps.sources import storages
from mayan.apps.storage.classes import DefinedStorage
from mayan.apps.testing.tests.base import BaseTestCase

from ..literals import STORAGE_NAME_SOURCE_CACHE_FOLDER
from ..settings import setting_source_cache_storage_backend_arguments


class SourcesStorageSettingsTestCase(BaseTestCase):
    def tearDown(self):
        super().tearDown()
        importlib.reload(storages)

    def test_setting_source_cache_storage_backend_arguments_invalid_value(self):
        self._set_environment_variable(
            name='MAYAN_{}'.format(
                setting_source_cache_storage_backend_arguments.global_name
            ), value="invalid_value"
        )
        self.test_case_silenced_logger_new_level = logging.FATAL + 10
        self._silence_logger(name='mayan.apps.storage.classes')

        with self.assertRaises(expected_exception=TypeError) as assertion:
            importlib.reload(storages)
            DefinedStorage.get(
                name=STORAGE_NAME_SOURCE_CACHE_FOLDER
            ).get_storage_instance()

        self.assertTrue(
            'Unable to initialize' in str(assertion.exception)
        )
        self.assertTrue(
            'staging folder' in str(assertion.exception)
        )



from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import (
    event_tag_attached, event_tag_created, event_tag_edited,
    event_tag_removed
)
from ..links import link_tag_edit
from ..models import Tag
from ..permissions import (
    permission_tag_attach, permission_tag_create, permission_tag_delete,
    permission_tag_edit, permission_tag_remove, permission_tag_view
)

from .mixins import DocumentTagViewTestMixin, TagViewTestMixin


class DocumentTagViewTestCase(
    DocumentTagViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_document_tags_list_no_permission(self):
        self._create_test_tag(add_test_document=True)

        self._clear_events()

        response = self._request_test_document_tag_list_view()
        self.assertNotContains(
            response=response, text=str(self._test_tag), status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tags_list_with_document_access(self):
        self._create_test_tag(add_test_document=True)

        self.grant_access(
            obj=self._test_document, permission=permission_tag_view
        )

        self._clear_events()

        response = self._request_test_document_tag_list_view()
        self.assertNotContains(
            response=response, text=str(self._test_tag), status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tags_list_with_tag_access(self):
        self._create_test_tag(add_test_document=True)

        self.grant_access(obj=self._test_tag, permission=permission_tag_view)

        self._clear_events()

        response = self._request_test_document_tag_list_view()
        self.assertNotContains(
            response=response, text=str(self._test_tag), status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tags_list_with_full_access(self):
        self._create_test_tag(add_test_document=True)

        self.grant_access(
            obj=self._test_document, permission=permission_tag_view
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_view
        )

        self._clear_events()

        response = self._request_test_document_tag_list_view()
        self.assertContains(
            response=response, text=str(self._test_tag),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_tags_list_with_full_access(self):
        self._create_test_tag(add_test_document=True)

        self.grant_access(
            obj=self._test_document, permission=permission_tag_view
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_tag_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tags_list_tag_edit_link_with_full_access(self):
        # Ensure that `DocumentTag` instances and links are
        # resolved in this view and not base Tag instances.
        self._create_test_tag(add_test_document=True)

        self.grant_access(
            obj=self._test_document, permission=permission_tag_view
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_edit
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_view
        )

        self._clear_events()

        response = self._request_test_document_tag_list_view()

        # Get the first context item that correspond to an an item in the
        # response nexted context list.
        for context in response.context:
            if 'object' in context:
                break

        context['object'] = self._test_tag

        result = link_tag_edit.resolve(context=context)

        self.assertNotContains(
            response=response, text=result.url,
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_attach_view_no_permission(self):
        self._create_test_tag()

        self._clear_events()

        response = self._request_test_document_tag_attach_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_attach_view_with_tag_access(self):
        self._create_test_tag()

        self.grant_access(
            obj=self._test_tag, permission=permission_tag_attach
        )

        self._clear_events()

        response = self._request_test_document_tag_attach_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_attach_view_with_document_access(self):
        self._create_test_tag()

        self.grant_access(
            obj=self._test_document, permission=permission_tag_attach
        )

        self._clear_events()

        response = self._request_test_document_tag_attach_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_attach_view_with_full_access(self):
        self._create_test_tag()

        self.grant_access(
            obj=self._test_document, permission=permission_tag_attach
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_attach
        )

        self._clear_events()

        response = self._request_test_document_tag_attach_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_tag)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_tag_attached.id)

    def test_trashed_document_tag_attach_view_with_full_access(self):
        self._create_test_tag()

        self.grant_access(
            obj=self._test_document, permission=permission_tag_attach
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_attach
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_tag_attach_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_tag_attach_view_no_permission(self):
        self._create_test_tag()

        self._clear_events()

        response = self._request_test_document_multiple_tag_multiple_attach_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_tag_attach_view_with_tag_access(self):
        self._create_test_tag()

        self.grant_access(
            obj=self._test_tag, permission=permission_tag_attach
        )

        self._clear_events()

        response = self._request_test_document_multiple_tag_multiple_attach_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_tag_attach_view_with_document_access(self):
        self._create_test_tag()

        self.grant_access(
            obj=self._test_document, permission=permission_tag_attach
        )

        self._clear_events()

        response = self._request_test_document_multiple_tag_multiple_attach_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_tag_attach_view_with_full_access(self):
        self._create_test_tag()

        self.grant_access(
            obj=self._test_document, permission=permission_tag_attach
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_attach
        )

        self._clear_events()

        response = self._request_test_document_multiple_tag_multiple_attach_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_tag)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_tag_attached.id)

    def test_trashed_document_multiple_tag_attach_view_with_full_access(self):
        self._create_test_tag()

        self.grant_access(
            obj=self._test_document, permission=permission_tag_attach
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_attach
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_multiple_tag_multiple_attach_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_multiple_remove_view_no_permission(self):
        self._create_test_tag()
        self._test_document.tags.add(self._test_tag)

        self._clear_events()

        response = self._request_test_document_tag_multiple_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_multiple_remove_view_with_tag_access(self):
        self._create_test_tag()
        self._test_document.tags.add(self._test_tag)

        self.grant_access(
            obj=self._test_tag, permission=permission_tag_remove
        )

        self._clear_events()

        response = self._request_test_document_tag_multiple_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_multiple_remove_view_with_document_access(self):
        self._create_test_tag()
        self._test_document.tags.add(self._test_tag)

        self.grant_access(
            obj=self._test_document, permission=permission_tag_remove
        )

        self._clear_events()

        response = self._request_test_document_tag_multiple_remove_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_tag_multiple_remove_view_with_full_access(self):
        self._create_test_tag()
        self._test_document.tags.add(self._test_tag)

        self.grant_access(
            obj=self._test_document, permission=permission_tag_remove
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_remove
        )

        self._clear_events()

        response = self._request_test_document_tag_multiple_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_tag)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_tag_removed.id)

    def test_trashed_document_tag_multiple_remove_view_with_full_access(self):
        self._create_test_tag()
        self._test_document.tags.add(self._test_tag)

        self.grant_access(
            obj=self._test_document, permission=permission_tag_remove
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_remove
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_tag_multiple_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_tag_multiple_remove_view_no_permission(self):
        self._create_test_tag()
        self._test_document.tags.add(self._test_tag)

        self._clear_events()

        response = self._request_test_document_multiple_tag_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_tag_remove_view_with_tag_access(self):
        self._create_test_tag()
        self._test_document.tags.add(self._test_tag)

        self.grant_access(
            obj=self._test_tag, permission=permission_tag_remove
        )

        self._clear_events()

        response = self._request_test_document_multiple_tag_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_tag_remove_view_with_document_access(self):
        self._create_test_tag()
        self._test_document.tags.add(self._test_tag)

        self.grant_access(
            obj=self._test_document, permission=permission_tag_remove
        )

        self._clear_events()

        response = self._request_test_document_multiple_tag_remove_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_multiple_tag_remove_view_with_full_access(self):
        self._create_test_tag()
        self._test_document.tags.add(self._test_tag)

        self.grant_access(
            obj=self._test_document, permission=permission_tag_remove
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_remove
        )

        self._clear_events()

        response = self._request_test_document_multiple_tag_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self._test_tag not in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_tag)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_tag_removed.id)

    def test_trashed_document_multiple_tag_remove_view_with_full_access(self):
        self._create_test_tag()
        self._test_document.tags.add(self._test_tag)

        self.grant_access(
            obj=self._test_document, permission=permission_tag_remove
        )
        self.grant_access(
            obj=self._test_tag, permission=permission_tag_remove
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_multiple_tag_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self._test_tag in self._test_document.tags.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_document_list_with_no_permission(self):
        self._create_test_tag(add_test_document=True)

        self._clear_events()

        response = self._request_test_tag_document_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_document_list_with_tag_access(self):
        self._create_test_tag(add_test_document=True)

        self.grant_access(obj=self._test_tag, permission=permission_tag_view)

        self._clear_events()

        response = self._request_test_tag_document_list_view()
        self.assertContains(
            response=response, text=str(self._test_tag),
            status_code=200
        )
        self.assertNotContains(
            response=response, text=str(self._test_document),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_document_list_with_document_access(self):
        self._create_test_tag(add_test_document=True)

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_tag_document_list_view()
        self.assertNotContains(
            response=response, text=str(self._test_tag),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self._test_document),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_document_list_with_full_access(self):
        self._create_test_tag(add_test_document=True)

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )
        self.grant_access(obj=self._test_tag, permission=permission_tag_view)

        self._clear_events()

        response = self._request_test_tag_document_list_view()
        self.assertContains(
            response=response, text=str(self._test_tag),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self._test_document),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_trashed_document_list_with_full_access(self):
        self._create_test_tag(add_test_document=True)

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )
        self.grant_access(obj=self._test_tag, permission=permission_tag_view)

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_tag_document_list_view()
        self.assertContains(
            response=response, text=str(self._test_tag),
            status_code=200
        )
        self.assertNotContains(
            response=response, text=str(self._test_document),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class TagViewTestCase(TagViewTestMixin, GenericViewTestCase):
    def test_tag_create_view_no_permission(self):
        tag_count = Tag.objects.count()

        self._clear_events()

        response = self._request_test_tag_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Tag.objects.count(), tag_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_create_view_with_permissions(self):
        self.grant_permission(permission=permission_tag_create)

        tag_count = Tag.objects.count()

        self._clear_events()

        response = self._request_test_tag_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Tag.objects.count(), tag_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_tag)
        self.assertEqual(events[0].verb, event_tag_created.id)

    def test_tag_delete_view_no_permission(self):
        self._create_test_tag()

        tag_count = Tag.objects.count()

        self._clear_events()

        response = self._request_test_tag_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Tag.objects.count(), tag_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_delete_view_with_access(self):
        self._create_test_tag()

        self.grant_access(
            obj=self._test_tag, permission=permission_tag_delete
        )

        tag_count = Tag.objects.count()

        self._clear_events()

        response = self._request_test_tag_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Tag.objects.count(), tag_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_multiple_delete_view_no_permission(self):
        self._create_test_tag()

        tag_count = Tag.objects.count()

        self._clear_events()

        response = self._request_test_tag_multiple_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Tag.objects.count(), tag_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_multiple_delete_view_with_access(self):
        self._create_test_tag()

        self.grant_access(
            obj=self._test_tag, permission=permission_tag_delete
        )

        tag_count = Tag.objects.count()

        self._clear_events()

        response = self._request_test_tag_multiple_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Tag.objects.count(), tag_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_edit_view_no_permission(self):
        self._create_test_tag()

        tag_label = self._test_tag.label

        self._clear_events()

        response = self._request_test_tag_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_tag.refresh_from_db()
        self.assertEqual(self._test_tag.label, tag_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_edit_view_with_access(self):
        self._create_test_tag()

        self.grant_access(obj=self._test_tag, permission=permission_tag_edit)

        tag_label = self._test_tag.label

        self._clear_events()

        response = self._request_test_tag_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_tag.refresh_from_db()
        self.assertNotEqual(self._test_tag.label, tag_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_tag)
        self.assertEqual(events[0].verb, event_tag_edited.id)

    def test_tag_list_view_with_no_permission(self):
        self._create_test_tag()

        self._clear_events()

        response = self._request_test_tag_list_view()
        self.assertNotContains(
            response=response, text=self._test_tag.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_tag_list_view_with_access(self):
        self._create_test_tag()

        self.grant_access(obj=self._test_tag, permission=permission_tag_view)

        self._clear_events()

        response = self._request_test_tag_list_view()
        self.assertContains(
            response=response, text=self._test_tag.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



TEST_SOURCE_BACKEND_PATH_INTERACTIVE = 'mayan.apps.source_interactive.tests.source_backends.SourceBackendTestInteractive'
TEST_SOURCE_BACKEND_PATH_INTERACTIVE_ACTION = 'mayan.apps.source_interactive.tests.source_backends.SourceBackendTestInteractiveAction'



from mayan.apps.source_web_forms.tests.mixins import WebFormSourceTestMixin
from mayan.apps.sources.tests.mixins.source_view_mixins import SourceActionViewTestMixin

from .metadata_type_mixins import MetadataTypeTestMixin


class MetadataDocumentUploadWizardStepTestMixin(
    MetadataTypeTestMixin, WebFormSourceTestMixin, SourceActionViewTestMixin
):
    def _request_test_source_metadata_upload_post_view_with_metadata_types(
        self, metadata_value=None
    ):
        if metadata_value is None:
            value = 'test metadata value'
        else:
            value = metadata_value

        extra_data = {}
        for index, test_metadata_type in enumerate(self._test_metadata_type_list):
            extra_data[
                'metadata{}_metadata_type_id'.format(index)
            ] = test_metadata_type.pk
            extra_data[
                'metadata{}_value'.format(index)
            ] = value

        return self._request_test_source_document_upload_post_view(
            extra_data=extra_data
        )



from ..classes import FileMetadataDriver

from .literals import (
    TEST_DRIVER_LABEL, TEST_DRIVER_INTERNAL_NAME, TEST_FILE_METADATA_KEY,
    TEST_FILE_METADATA_VALUE
)


class FileMetadataDriverTest(FileMetadataDriver):
    internal_name = TEST_DRIVER_INTERNAL_NAME
    label = TEST_DRIVER_LABEL
    mime_type_list = ('*',)

    def _process(self, document_file):
        return {
            TEST_FILE_METADATA_KEY: TEST_FILE_METADATA_VALUE
        }



from mayan.apps.appearance.classes import Icon

icon_impersonate_start = Icon(
    driver_name='fontawesome-dual', primary_symbol='user',
    secondary_symbol='user'
)
icon_login = Icon(driver_name='fontawesome', symbol='sign-in-alt')
icon_logout = Icon(driver_name='fontawesome', symbol='sign-out-alt')
icon_password_change = Icon(driver_name='fontawesome', symbol='key')



from django.db import migrations


def code_copy_messages(apps, schema_editor):
    AccessControlList = apps.get_model(
        app_label='acls', model_name='AccessControlList'
    )
    Action = apps.get_model(
        app_label='actstream', model_name='Action'
    )
    Announcement = apps.get_model(
        app_label='announcements', model_name='Announcement'
    )
    ContentType = apps.get_model(
        app_label='contenttypes', model_name='ContentType'
    )
    Message = apps.get_model(app_label='motd', model_name='Message')
    StoredEventType = apps.get_model(
        app_label='events', model_name='StoredEventType'
    )
    StoredPermission = apps.get_model(
        app_label='permissions', model_name='StoredPermission'
    )

    announcement_content_type = ContentType.objects.get_for_model(
        model=Announcement
    )
    message_content_type = ContentType.objects.get_for_model(model=Message)

    # Copy the messages.
    for message in Message.objects.using(alias=schema_editor.connection.alias).all():
        announcement = Announcement.objects.create(
            enabled=message.enabled, label=message.label,
            text=message.message, start_datetime=message.start_datetime,
            end_datetime=message.end_datetime
        )

        queryset_message_acls = AccessControlList.objects.using(
            alias=schema_editor.connection.alias
        ).filter(
            content_type=message_content_type, object_id=message.pk
        )

        # Update the ACLs.
        for message_acl in queryset_message_acls:
            message_acl._event_ignore = True
            message_acl.content_type = announcement_content_type
            message_acl.object_id = announcement.pk
            message_acl.save()

        queryset_message_action = Action.objects.using(
            alias=schema_editor.connection.alias
        ).filter(
            target_content_type=message_content_type,
            target_object_id=message.pk
        )

        # Update the actions.
        for message_action in queryset_message_action:
            message_action.target_content_type = announcement_content_type
            message_action.target_object_id = announcement.pk
            message_action.save()

        message.delete()

    queryset_stored_event = StoredEventType.objects.using(
        alias=schema_editor.connection.alias
    ).filter(name__startswith='motd')

    # Update the stored events of the MOTD app.
    for stored_event_type in queryset_stored_event:
        stored_event_type.name.replace('motd', 'announcements').replace('message', 'announcement')
        stored_event_type.save()

    queryset_stored_permissions = StoredPermission.objects.using(
        alias=schema_editor.connection.alias
    ).filter(namespace='motd')

    # Update the stored permissions of the MOTD app.
    for stored_permission in queryset_stored_permissions:
        stored_permission.namespace.replace('motd', 'announcements')
        stored_permission.name.replace('message', 'announcement')
        stored_permission.save()


def code_copy_messages_reverse(apps, schema_editor):
    AccessControlList = apps.get_model(
        app_label='acls', model_name='AccessControlList'
    )
    Action = apps.get_model(
        app_label='actstream', model_name='Action'
    )
    Announcement = apps.get_model(
        app_label='announcements', model_name='Announcement'
    )
    ContentType = apps.get_model(
        app_label='contenttypes', model_name='ContentType'
    )
    Message = apps.get_model(app_label='motd', model_name='Message')
    StoredEventType = apps.get_model(
        app_label='events', model_name='StoredEventType'
    )
    StoredPermission = apps.get_model(
        app_label='permissions', model_name='StoredPermission'
    )

    announcement_content_type = ContentType.objects.get_for_model(
        model=Announcement
    )

    message_content_type = ContentType.objects.get_for_model(model=Message)

    # Copy the announcements.
    for announcement in Announcement.objects.using(alias=schema_editor.connection.alias).all():
        message = Message.objects.create(
            enabled=announcement.enabled, label=announcement.label,
            message=announcement.text,
            start_datetime=announcement.start_datetime,
            end_datetime=announcement.end_datetime
        )

        queryset_announcement_acls = AccessControlList.objects.using(
            alias=schema_editor.connection.alias
        ).filter(
            content_type=announcement_content_type, object_id=announcement.pk
        )

        # Update the ACLs.
        for announcement_acl in queryset_announcement_acls:
            announcement_acl._event_ignore = True
            announcement_acl.content_type = message_content_type
            announcement_acl.object_id = message.pk
            announcement_acl.save()

        queryset_announcement_action = Action.objects.using(
            alias=schema_editor.connection.alias
        ).filter(
            target_content_type=announcement_content_type,
            target_object_id=announcement.pk
        )

        # Update the actions.
        for announcement_action in queryset_announcement_action:
            announcement_action.target_content_type = message_content_type
            announcement_action.target_object_id = message.pk
            announcement_action.save()

        announcement.delete()

    queryset_stored_event = StoredEventType.objects.using(
        alias=schema_editor.connection.alias
    ).filter(name__startswith='motd')

    # Update the stored events of the announcements app.
    for stored_event_type in queryset_stored_event:
        stored_event_type.name.replace('announcements', 'motd').replace('announcement', 'message')
        stored_event_type.save()

    queryset_stored_permissions = StoredPermission.objects.using(
        alias=schema_editor.connection.alias
    ).filter(namespace='announcements')

    # Update the stored permissions of the announcements app.
    for stored_permission in queryset_stored_permissions:
        stored_permission.namespace.replace('announcements', 'motd')
        stored_permission.name.replace('announcement', 'message')
        stored_permission.save()


class Migration(migrations.Migration):
    dependencies = [
        ('acls', '0004_auto_20210130_0322'),
        ('announcements', '0001_initial'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('events', '0008_auto_20180315_0029'),
        ('permissions', '0004_auto_20191213_0044')
    ]

    operations = [
        migrations.RunPython(
            code=code_copy_messages,
            reverse_code=code_copy_messages_reverse
        )
    ]



import functools
import logging
from pathlib import Path

import whoosh
from whoosh import qparser
from whoosh.filedb.filestore import FileStorage
from whoosh.index import EmptyIndexError
from whoosh.qparser import (
    FuzzyTermPlugin, GtLtPlugin, MultifieldParser, OrGroup, RegexPlugin
)
from whoosh.qparser.dateparse import DateParserPlugin
from whoosh.query import Every
from whoosh.writing import BufferedWriter

from django.conf import settings

from mayan.apps.common.utils import any_to_bool
from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.exceptions import LockError
from mayan.apps.storage.utils import TemporaryDirectory

from ..exceptions import (
    DynamicSearchBackendException, DynamicSearchRetry,
    DynamicSearchValueTransformationError
)
from ..search_backends import SearchBackend
from ..search_fields import SearchFieldVirtualAllFields
from ..search_models import SearchModel
from ..search_query_types import (
    BackendQueryType, QueryTypeExact, QueryTypeFuzzy, QueryTypeGreaterThan,
    QueryTypeGreaterThanOrEqual, QueryTypeLessThan,
    QueryTypeLessThanOrEqual, QueryTypePartial, QueryTypeRange,
    QueryTypeRangeExclusive, QueryTypeRegularExpression
)

from .literals.whoosh_literals import (
    DJANGO_TO_WHOOSH_FIELD_MAP, TEXT_LOCK_INSTANCE_DEINDEX,
    TEXT_LOCK_INSTANCE_INDEX, WHOOSH_INDEX_DIRECTORY_NAME
)

logger = logging.getLogger(name=__name__)


class BackendQueryTypeWhoosh(BackendQueryType):
    def do_resolve(self):
        if self.get_search_backend_field_type() == whoosh.fields.DATETIME:

            self.extra_kwargs['parser'].add_plugin(
                DateParserPlugin()
            )

        return self._do_resolve()


class BackendQueryTypeExact(BackendQueryTypeWhoosh):
    query_type = QueryTypeExact

    def _do_resolve(self):
        if self.value is not None:
            if not self.value and self.is_quoted_value:
                return 'NOT *'
            else:
                if self.is_quoted_value:
                    template = '{}:("{}")'
                else:
                    template = '{}:({})'

                return template.format(
                    self.search_field.field_name, self.value
                )


class BackendQueryTypeFuzzy(BackendQueryTypeWhoosh):
    query_type = QueryTypeFuzzy

    def _do_resolve(self):
        self.extra_kwargs['parser'].add_plugin(
            FuzzyTermPlugin()
        )

        if self.value is not None:
            if self.is_quoted_value:
                template = '{}:"{}"~2'
            else:
                template = '{}:{}~2'

            return template.format(
                self.search_field.field_name, self.value
            )


class BackendQueryTypeComparison(BackendQueryTypeWhoosh):
    def _do_resolve(self):
        self.extra_kwargs['parser'].add_plugin(
            GtLtPlugin()
        )

        if self.value is not None:
            return self.template.format(
                self.search_field.field_name, self.value
            )


class BackendQueryTypeGreaterThan(BackendQueryTypeComparison):
    query_type = QueryTypeGreaterThan
    template = '{}:>{}'


class BackendQueryTypeGreaterThanOrEqual(BackendQueryTypeComparison):
    query_type = QueryTypeGreaterThanOrEqual
    template = '{}:>={}'


class BackendQueryTypeLessThan(BackendQueryTypeComparison):
    query_type = QueryTypeLessThan
    template = '{}:<{}'


class BackendQueryTypeLessThanOrEqual(BackendQueryTypeComparison):
    query_type = QueryTypeLessThanOrEqual
    template = '{}:<={}'


class BackendQueryTypePartial(BackendQueryTypeWhoosh):
    query_type = QueryTypePartial

    def _do_resolve(self):
        if self.value is not None:
            if self.is_quoted_value:
                template = '{}:"{}"'
            else:
                template = '{}:(*{}*)'

            if self.get_search_backend_field_type() == whoosh.fields.BOOLEAN:
                template = '{}:({})'

            if self.value is not None:
                return template.format(
                    self.search_field.field_name, self.value
                )


class BackendQueryTypeRange(BackendQueryTypeWhoosh):
    query_type = QueryTypeRange

    def _do_resolve(self):
        if self.value is not None:
            return '{}:[{} TO {}]'.format(
                self.search_field.field_name, *self.value
            )


class BackendQueryTypeRangeExclusive(BackendQueryTypeWhoosh):
    query_type = QueryTypeRangeExclusive

    def _do_resolve(self):
        if self.value is not None:
            return '{}:{{{} TO {}}}'.format(
                self.search_field.field_name, *self.value
            )


class BackendQueryTypeRegularExpression(BackendQueryTypeWhoosh):
    query_type = QueryTypeRegularExpression

    def _do_resolve(self):
        self.extra_kwargs['parser'].add_plugin(
            RegexPlugin()
        )

        if self.value is not None:
            return '{}:r"{}"'.format(
                self.search_field.field_name, self.value
            )


class WhooshSearchBackend(SearchBackend):
    _local_attribute_backend_temporary_directory = None
    feature_reindex = True
    field_type_mapping = DJANGO_TO_WHOOSH_FIELD_MAP

    def __init__(
        self, index_path=None, writer_limitmb=128, writer_multisegment=False,
        writer_procs=1, **kwargs
    ):
        super().__init__(**kwargs)

        if self._test_mode:
            if not self.__class__._local_attribute_backend_temporary_directory:
                self.__class__._local_attribute_backend_temporary_directory = TemporaryDirectory()
            index_path = self._local_attribute_backend_temporary_directory.name

        self.index_path = Path(
            index_path or Path(
                settings.MEDIA_ROOT, WHOOSH_INDEX_DIRECTORY_NAME
            )
        )

        if writer_limitmb:
            writer_limitmb = int(writer_limitmb)

        if writer_multisegment:
            writer_multisegment = any_to_bool(value=writer_multisegment)

        if writer_procs:
            writer_procs = int(writer_procs)

        self.writer_kwargs = {
            'limitmb': writer_limitmb, 'multisegment': writer_multisegment,
            'procs': writer_procs
        }

    def _clear_search_model_index(self, search_model):
        schema = self._get_search_model_schema(search_model=search_model)

        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            # Clear the model index.
            self._get_storage().create_index(
                indexname=search_model.full_name, schema=schema
            )

    def _do_query_resolve(self, index, limit, query):
        with index.searcher() as searcher:
            results = searcher.search(q=query, limit=limit)
            logger.debug('results: %s', results)
            return [
                int(
                    result['id']
                ) for result in results
            ]

    def _get_or_create_index(self, search_model):
        storage = self._get_storage()
        schema = self._get_search_model_schema(search_model=search_model)

        try:
            # Explicitly specify the schema. Allows using existing index
            # when the schema changes.
            index = storage.open_index(
                indexname=search_model.full_name, schema=schema
            )
        except EmptyIndexError:
            index = storage.create_index(
                indexname=search_model.full_name, schema=schema
            )

        return index

    @functools.cache
    def _get_search_model_schema(self, search_model):
        field_map = self.get_resolved_field_type_map(
            search_model=search_model
        )
        schema_kwargs = {
            key: value['field'] for key, value in field_map.items()
        }

        return whoosh.fields.Schema(**schema_kwargs)

    def _get_status(self):
        result = []

        title = 'Whoosh search model indexing status'
        result.append(title)
        result.append(len(title) * '=')

        for search_model in SearchModel.all():
            index = self._get_or_create_index(search_model=search_model)
            search_results = index.searcher().search(
                q=Every('id')
            )

            result.append(
                '{}: {}'.format(
                    search_model.label, search_results.estimated_length()
                )
            )

        return '\n'.join(result)

    def _get_storage(self):
        return FileStorage(path=self.index_path)

    def _get_writer(self, search_model):
        index = self._get_or_create_index(search_model=search_model)

        return index.writer(**self.writer_kwargs)

    def _initialize(self):
        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            self.index_path.mkdir(exist_ok=True)

    def _search(
        self, limit, search_field, query_type, value, is_quoted_value=False,
        is_raw_value=False
    ):
        self.do_query_type_verify(
            query_type=query_type, search_field=search_field
        )

        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            index = self._get_or_create_index(
                search_model=search_field.search_model
            )

            if isinstance(search_field, SearchFieldVirtualAllFields):
                parser = MultifieldParser(
                    [
                        search_field.field_name for search_field in search_field.field_composition
                    ], schema=index.schema, group=OrGroup
                )
                search_field_queries = []

                for search_field in search_field.field_composition:
                    try:
                        search_string = query_type.resolve_for_backend(
                            is_quoted_value=is_quoted_value,
                            is_raw_value=is_raw_value, search_backend=self,
                            search_field=search_field, value=value,
                            extra_kwargs={
                                'parser': parser
                            }
                        )
                    except DynamicSearchValueTransformationError:
                        """Skip the search field."""
                    else:
                        if search_string is not None:
                            search_field_queries.append(search_string)

                query = parser.parse(
                    ' '.join(search_field_queries)
                )
            else:
                parser = qparser.QueryParser(
                    fieldname=search_field.field_name, schema=index.schema
                )

                try:
                    search_string = query_type.resolve_for_backend(
                        is_quoted_value=is_quoted_value,
                        is_raw_value=is_raw_value, search_backend=self,
                        search_field=search_field, value=value,
                        extra_kwargs={
                            'parser': parser
                        }
                    )
                except DynamicSearchValueTransformationError:
                    return ()
                else:
                    if search_string is None:
                        return ()

                logger.debug('search_string: %s', search_string)

                query = parser.parse(text=search_string)

            return self._do_query_resolve(
                index=index, limit=limit, query=query
            )
        else:
            return ()

    def _update_mappings(self, search_model=None):
        if search_model:
            search_models = (search_model,)
        else:
            search_models = SearchModel.all()

        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            for search_model in search_models:
                self._get_or_create_index(search_model=search_model)

    def deindex_instance(self, instance):
        try:
            lock = LockingBackend.get_backend().acquire_lock(
                name=TEXT_LOCK_INSTANCE_DEINDEX
            )
        except LockError:
            raise
        else:
            try:
                search_model = SearchModel.get_for_model(instance=instance)
                index = self._get_or_create_index(search_model=search_model)

                if not settings.COMMON_DISABLE_LOCAL_STORAGE:
                    with index.writer(**self.writer_kwargs) as writer:
                        writer.delete_by_term(
                            'id', str(instance.pk)
                        )
            finally:
                lock.release()

    def do_native_type_conversion(self, value):
        if isinstance(value, (list, tuple)):
            return ' '.join(value)
        else:
            return value

    def index_instance(self, instance, exclude_model=None, exclude_kwargs=None):
        try:
            lock = LockingBackend.get_backend().acquire_lock(
                name=TEXT_LOCK_INSTANCE_INDEX
            )
        except LockError:
            raise
        else:
            try:
                search_model = SearchModel.get_for_model(instance=instance)
                if not settings.COMMON_DISABLE_LOCAL_STORAGE:
                    with self._get_writer(search_model=search_model) as writer:
                        try:
                            writer.delete_by_term(
                                'id', str(instance.pk)
                            )
                        except Exception as exception:
                            # The parenthesis is used to define a multi
                            # line error message not a translatable string.
                            error_text = (
                                'Unexpected exception while '
                                'deleting search object id: {id}, '
                                'search model: {search_model}, '
                                'raw data: {raw_data}, '
                                'field map: {field_map}; '
                                '{exception}'
                            ).format(
                                exception=exception,
                                field_map=self.get_resolved_field_type_map(
                                    search_model=search_model
                                ), id=instance.pk,
                                raw_data=instance.__dict__,
                                search_model=search_model.full_name
                            )

                            logger.error(error_text, exc_info=True)
                            raise DynamicSearchBackendException(
                                error_text
                            ) from exception
                        else:
                            kwargs = search_model.populate(
                                search_backend=self, instance=instance,
                                exclude_model=exclude_model,
                                exclude_kwargs=exclude_kwargs
                            )

                            try:
                                writer.add_document(**kwargs)
                            except Exception as exception:
                                # The parenthesis is used to define a multi
                                # line error message not a translatable
                                # string.
                                error_text = (
                                    'Unexpected exception while '
                                    'indexing search object id: {id}, '
                                    'search model: {search_model}, '
                                    'index data: {index_data}, '
                                    'raw data: {raw_data}, '
                                    'field map: {field_map}; '
                                    '{exception}'
                                ).format(
                                    exception=exception,
                                    field_map=self.get_resolved_field_type_map(
                                        search_model=search_model
                                    ), id=instance.pk, index_data=kwargs,
                                    raw_data=instance.__dict__,
                                    search_model=search_model.full_name
                                )

                                logger.error(error_text, exc_info=True)
                                raise DynamicSearchBackendException(
                                    error_text
                                ) from exception

            except whoosh.index.LockError:
                raise DynamicSearchRetry
            finally:
                lock.release()

    def index_instances(self, search_model, id_list):
        queryset = search_model.get_queryset()
        queryset = queryset.filter(pk__in=id_list)

        try:
            lock = LockingBackend.get_backend().acquire_lock(
                name=TEXT_LOCK_INSTANCE_INDEX
            )
        except LockError:
            raise
        else:
            try:
                if not settings.COMMON_DISABLE_LOCAL_STORAGE:
                    index = self._get_or_create_index(search_model=search_model)

                    writer = BufferedWriter(index=index)
                    for instance in queryset:
                        kwargs = search_model.populate(
                            search_backend=self, instance=instance
                        )

                        try:
                            writer.update_document(**kwargs)
                        except Exception as exception:
                            # The parenthesis is used to define a multi
                            # line error message not a translatable string.
                            error_text = (
                                'Unexpected exception while '
                                'indexing search model: {search_model}, '
                                'id_list: {id_list}',
                                'index data: {index_data}, '
                                'raw data: {raw_data}, '
                                'field map: {field_map}; '
                                '{exception}'
                            ).format(
                                exception=exception,
                                field_map=self.get_resolved_field_type_map(
                                    search_model=search_model
                                ), id_list=id_list, index_data=kwargs,
                                raw_data=instance.__dict__,
                                search_model=search_model.full_name
                            )

                            logger.error(error_text, exc_info=True)
                            raise DynamicSearchBackendException(
                                error_text
                            ) from exception
                    writer.close()
            except whoosh.index.LockError:
                raise DynamicSearchRetry
            finally:
                lock.release()

    def reset(self, search_model=None):
        self.tear_down(search_model=search_model)
        self._update_mappings(search_model=search_model)

    def tear_down(self, search_model=None):
        if search_model:
            search_models = (search_model,)
        else:
            search_models = SearchModel.all()

        if not settings.COMMON_DISABLE_LOCAL_STORAGE:
            for search_model in search_models:
                self._clear_search_model_index(search_model=search_model)

    def test_mode_stop(self):
        self.__class__._local_attribute_backend_temporary_directory.cleanup()


BackendQueryType.register(
    klass=BackendQueryTypeExact, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeFuzzy, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypePartial, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeGreaterThan, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeGreaterThanOrEqual,
    search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeLessThan, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeLessThanOrEqual, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRange, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRangeExclusive, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRegularExpression,
    search_backend=WhooshSearchBackend
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Key management'), name='django_gpg'
)

event_key_created = namespace.add_event_type(
    label=_(message='Key created'),
    name='key_created'
)
event_key_downloaded = namespace.add_event_type(
    label=_(message='Key downloaded'), name='key_downloaded'
)



import os

from django.conf import settings
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.literals import TIME_DELTA_UNIT_DAYS

CHECK_DELETE_PERIOD_INTERVAL = 60
CHECK_TRASH_PERIOD_INTERVAL = 60

DEFAULT_DELETE_PERIOD = 30
DEFAULT_DELETE_TIME_UNIT = TIME_DELTA_UNIT_DAYS
DEFAULT_DOCUMENT_FILE_ACTION_NAME = 'replace'
DEFAULT_DOCUMENT_TYPE_LABEL = _(message='Default')

# Old defaults (<4.0), used for the setting migrations.
DEFAULT_DOCUMENTS_STORAGE_BACKEND = 'django.core.files.storage.FileSystemStorage'
DEFAULT_DOCUMENTS_STORAGE_BACKEND_ARGUMENTS = {
    'location': os.path.join(settings.MEDIA_ROOT, 'document_storage')
}

DEFAULT_DOCUMENTS_DISPLAY_HEIGHT = ''
DEFAULT_DOCUMENTS_DISPLAY_WIDTH = '3600'
DEFAULT_DOCUMENTS_FAVORITE_COUNT = 400
DEFAULT_DOCUMENTS_FILE_PAGE_IMAGE_CACHE_MAXIMUM_SIZE = 500 * 2 ** 20  # 500 Megabytes
DEFAULT_DOCUMENTS_FILE_STORAGE_BACKEND = 'django.core.files.storage.FileSystemStorage'
DEFAULT_DOCUMENTS_FILE_STORAGE_BACKEND_ARGUMENTS = {
    'location': os.path.join(settings.MEDIA_ROOT, 'document_file_storage')
}
DEFAULT_DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND = 'django.core.files.storage.FileSystemStorage'
DEFAULT_DOCUMENTS_FILE_PAGE_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS = {
    'location': os.path.join(
        settings.MEDIA_ROOT, 'document_file_page_image_cache'
    )
}
DEFAULT_DOCUMENTS_HASH_BLOCK_SIZE = 65535
DEFAULT_DOCUMENTS_LIST_THUMBNAIL_WIDTH = '50'
DEFAULT_DOCUMENTS_PREVIEW_HEIGHT = ''
DEFAULT_DOCUMENTS_PREVIEW_WIDTH = '1600'
DEFAULT_DOCUMENTS_PRINT_HEIGHT = ''
DEFAULT_DOCUMENTS_PRINT_WIDTH = '3600'
DEFAULT_DOCUMENTS_RECENTLY_ACCESSED_COUNT = 400
DEFAULT_DOCUMENTS_RECENTLY_CREATED_COUNT = 400
DEFAULT_DOCUMENTS_ROTATION_STEP = 90
DEFAULT_DOCUMENTS_THUMBNAIL_HEIGHT = ''
DEFAULT_DOCUMENTS_THUMBNAIL_WIDTH = '800'
DEFAULT_DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_MAXIMUM_SIZE = 500 * 2 ** 20  # 500 Megabytes
DEFAULT_DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND = 'django.core.files.storage.FileSystemStorage'
DEFAULT_DOCUMENTS_VERSION_PAGE_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS = {
    'location': os.path.join(
        settings.MEDIA_ROOT, 'document_version_page_image_cache'
    )
}

DEFAULT_DOCUMENTS_ZOOM_MAX_LEVEL = 300
DEFAULT_DOCUMENTS_ZOOM_MIN_LEVEL = 25
DEFAULT_DOCUMENTS_ZOOM_PERCENT_STEP = 25
DEFAULT_LANGUAGE = 'eng'
DEFAULT_LANGUAGE_CODES = (
    'ilo', 'run', 'uig', 'hin', 'pan', 'pnb', 'wuu', 'msa', 'kxd', 'ind',
    'zsm', 'jax', 'meo', 'kvr', 'xmm', 'min', 'mui', 'zmi', 'max', 'mfa',
    'cjy', 'nan', 'pus', 'pbu', 'pbt', 'wne', 'hsn', 'hak', 'ful', 'fuc',
    'fuf', 'ffm', 'fue', 'fuh', 'fuq', 'fuv', 'fub', 'fui', 'nep', 'npi',
    'dty', 'sin', 'khm', 'kxm', 'ell', 'grc', 'cpg', 'gmy', 'pnt', 'tsd',
    'yej', 'nya', 'mnp', 'dhd', 'cdo', 'hil', 'bcc', 'bgn', 'bgp', 'cmn',
    'kok', 'spa', 'eng', 'ara', 'por', 'ben', 'rus', 'jpn', 'deu', 'jav',
    'tel', 'vie', 'kor', 'fra', 'mar', 'tam', 'urd', 'tur', 'ita', 'yue',
    'tha', 'guj', 'fas', 'pol', 'kan', 'mal', 'sun', 'hau', 'ory', 'mya',
    'ukr', 'bho', 'tgl', 'yor', 'mai', 'uzb', 'snd', 'amh', 'ron', 'orm',
    'ibo', 'aze', 'awa', 'gan', 'ceb', 'nld', 'kur', 'hbs', 'mlg', 'skr',
    'ctg', 'zha', 'tuk', 'asm', 'mad', 'som', 'mwr', 'mag', 'bgc', 'hun',
    'hne', 'dcc', 'aka', 'kaz', 'syl', 'zul', 'ces', 'kin', 'hat', 'que',
    'swe', 'hmn', 'sna', 'mos', 'xho', 'bel', 'heb'
)
DEFAULT_DOCUMENT_STUB_EXPIRATION_INTERVAL = 60 * 60 * 24  # 24 hours

DOCUMENT_FILE_PAGE_CREATE_BATCH_SIZE = 100
DOCUMENT_VERSION_PAGE_CREATE_BATCH_SIZE = 100

IMAGE_ERROR_NO_ACTIVE_VERSION = 'document_no_active_version'
IMAGE_ERROR_NO_VERSION_PAGES = 'document_no_version_pages'
IMAGE_ERROR_FILE_PAGE_TRANSFORMATION_ERROR = 'document_file_page_transformation_error'
IMAGE_ERROR_VERSION_PAGE_TRANSFORMATION_ERROR = 'document_version_page_transformation_error'

INTERVAL_TASK_STUBS_DELETION = 60 * 10  # 10 minutes

MONTH_NAMES = (
    _(message='January'), _(message='February'), _(message='March'), _(message='April'), _(message='May'),
    _(message='June'), _(message='July'), _(message='August'), _(message='September'), _(message='October'),
    _(message='November'), _(message='December')
)

PAGE_RANGE_ALL = 'all'
PAGE_RANGE_RANGE = 'range'
PAGE_RANGE_CHOICES = (
    (PAGE_RANGE_ALL, _(message='All pages')), (PAGE_RANGE_RANGE, _(message='Page range'))
)

STORAGE_NAME_DOCUMENT_FILE_PAGE_IMAGE_CACHE = 'documents__documentfilepageimagecache'
STORAGE_NAME_DOCUMENT_FILES = 'documents__documentfiles'
STORAGE_NAME_DOCUMENT_VERSION_PAGE_IMAGE_CACHE = 'documents__documentversionpageimagecache'



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Cabinets'), name='cabinets'
)

# Translators: this refers to the permission that will allow users to add
# documents to cabinets.
permission_cabinet_add_document = namespace.add_permission(
    label=_(message='Add documents to cabinets'), name='cabinet_add_document'
)
permission_cabinet_create = namespace.add_permission(
    label=_(message='Create cabinets'), name='cabinet_create'
)
permission_cabinet_delete = namespace.add_permission(
    label=_(message='Delete cabinets'), name='cabinet_delete'
)
permission_cabinet_edit = namespace.add_permission(
    label=_(message='Edit cabinets'), name='cabinet_edit'
)
permission_cabinet_remove_document = namespace.add_permission(
    label=_(message='Remove documents from cabinets'), name='cabinet_remove_document'
)
permission_cabinet_view = namespace.add_permission(
    label=_(message='View cabinets'), name='cabinet_view'
)



import datetime

from django.conf import settings
from django.db.migrations import Migration


def patch_Migration():
    _method_original__apply = Migration.apply
    _method_original__unapply = Migration.unapply

    def _method_patched__apply(self, *args, **kwargs):
        now_start = datetime.datetime.now()

        result = _method_original__apply(self, *args, **kwargs)

        time_delta = datetime.datetime.now() - now_start
        print(
            ' (Time delta: {})'.format(time_delta), end=''
        )
        return result

    def _method_patched__unapply(self, *args, **kwargs):
        now_start = datetime.datetime.now()

        result = _method_original__unapply(self, *args, **kwargs)

        time_delta = datetime.datetime.now() - now_start
        print(
            ' (Time delta: {})'.format(time_delta), end=''
        )
        return result

    if settings.DEBUG:
        Migration.apply = _method_patched__apply
        Migration.unapply = _method_patched__unapply



from django.contrib.auth.models import Group
from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import (
    AddRemoveView, MultipleObjectDeleteView, SingleObjectCreateView,
    SingleObjectDetailView, SingleObjectEditView, SingleObjectListView
)

from ..icons import (
    icon_group_create, icon_group_detail, icon_group_edit, icon_group_list,
    icon_group_setup, icon_group_single_delete, icon_group_user_list
)
from ..links import link_group_create
from ..permissions import (
    permission_group_create, permission_group_delete, permission_group_edit,
    permission_group_view, permission_user_edit
)
from ..querysets import get_user_queryset


class GroupCreateView(SingleObjectCreateView):
    extra_context = {
        'title': _(message='Create new group')
    }
    fields = ('name',)
    model = Group
    post_action_redirect = reverse_lazy(
        viewname='user_management:group_list'
    )
    view_icon = icon_group_create
    view_permission = permission_group_create

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class GroupDeleteView(MultipleObjectDeleteView):
    error_message = _(message='Error deleting group "%(instance)s"; %(exception)s')
    model = Group
    object_permission = permission_group_delete
    pk_url_kwarg = 'group_id'
    post_action_redirect = reverse_lazy(
        viewname='user_management:group_list'
    )
    success_message_plural = _(message='%(count)d groups deleted successfully.')
    success_message_single = _(message='Group "%(object)s" deleted successfully.')
    success_message_singular = _(message='%(count)d group deleted successfully.')
    title_plural = _(message='Delete the %(count)d selected groups.')
    title_single = _(message='Delete group: %(object)s.')
    title_singular = _(message='Delete the %(count)d selected group.')
    view_icon = icon_group_single_delete


class GroupDetailView(SingleObjectDetailView):
    fields = ('name',)
    model = Group
    object_permission = permission_group_view
    pk_url_kwarg = 'group_id'
    view_icon = icon_group_detail

    def get_extra_context(self, **kwargs):
        return {
            'object': self.object,
            'title': _(message='Details of group: %s') % self.object
        }


class GroupEditView(SingleObjectEditView):
    fields = ('name',)
    model = Group
    object_permission = permission_group_edit
    pk_url_kwarg = 'group_id'
    post_action_redirect = reverse_lazy(
        viewname='user_management:group_list'
    )
    view_icon = icon_group_edit

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(message='Edit group: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class GroupListView(SingleObjectListView):
    model = Group
    object_permission = permission_group_view
    view_icon = icon_group_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_group_setup,
            'no_results_main_link': link_group_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'User groups are organizational units. They should '
                'mirror the organizational units of your organization. '
                'Groups can\'t be used for access control. Use roles '
                'for permissions and access control, add groups to '
                'them.'
            ),
            'no_results_title': _(message='There are no user groups'),
            'title': _(message='Groups')
        }


class GroupUserAddRemoveView(AddRemoveView):
    list_available_title = _(message='Available users')
    list_added_title = _(message='Group users')
    main_object_method_add_name = 'users_add'
    main_object_method_remove_name = 'users_remove'
    main_object_model = Group
    main_object_permission = permission_group_edit
    main_object_pk_url_kwarg = 'group_id'
    secondary_object_permission = permission_user_edit
    view_icon = icon_group_user_list

    def get_actions_extra_kwargs(self):
        return {'user': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'title': _(message='Users of group: %s') % self.main_object
        }

    def get_list_added_queryset(self):
        return self.main_object.get_users(
            permission=permission_user_edit, user=self.request.user
        )

    def get_secondary_object_source_queryset(self):
        return get_user_queryset(user=self.request.user)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('mailer', '0006_auto_20191213_0044')
    ]

    operations = [
        migrations.DeleteModel(name='LogEntry'),
        migrations.DeleteModel(name='UserMailerLogEntry')
    ]



from mayan.apps.views.http import URL

from .settings import (
    setting_organization_installation_url,
    setting_organization_url_base_path
)


def get_organization_installation_url(request=None):
    installation_url = setting_organization_installation_url.value
    installation_path = setting_organization_url_base_path.value

    if installation_url:
        return URL(
            path=installation_path, url=installation_url
        ).to_string()
    elif request:
        return URL(
            netloc=request.get_host(), path=installation_path,
            port=request.get_port(), scheme=request.scheme
        ).to_string()
    else:
        return ''



from django.urls import re_path

from .views import QueueTaskTypeListView, WorkerListView, WorkerQueueListView

urlpatterns = [
    re_path(
        route=r'^queues/(?P<queue_name>\w+)/task_types/$',
        view=QueueTaskTypeListView.as_view(), name='queue_task_type_list'
    ),
    re_path(
        route=r'^workers/$', view=WorkerListView.as_view(),
        name='worker_list'
    ),
    re_path(
        route=r'^workers/(?P<worker_name>\w+)/queues/$',
        view=WorkerQueueListView.as_view(), name='worker_queue_list'
    )
]



from mayan.apps.common.utils import get_class_full_name
from mayan.apps.sources.models import Source


class SourceBackendActionMixinCallbackBase:
    def _assemble_callback(self, callback_name, result, task_kwargs):
        callback_kwargs_base = self.get_callback_kwargs()

        callback_kwargs = callback_kwargs_base.copy()

        self_get_callback_kwargs_method_name = 'get_callback_kwargs_{}'.format(callback_name)
        self_get_callback_kwargs_method = getattr(
            self, self_get_callback_kwargs_method_name
        )

        callback_kwargs.update(
            **self_get_callback_kwargs_method(task_kwargs=task_kwargs)
        )

        callback_info = {
            'dotted_path': get_class_full_name(klass=Source),
            'function_name': 'callback_{}'.format(callback_name),
            'kwargs': callback_kwargs
        }

        result.setdefault(
            'callback_dict', {}
        )
        callback_dict = result['callback_dict']
        callback_dict.setdefault(
            callback_name, callback_info
        )

    def get_callback_kwargs(self, **kwargs):
        return {
            'source_id': self.source.pk
        }

    def get_callback_kwargs_post_document_create(self, task_kwargs):
        return {}

    def get_callback_kwargs_post_document_file_create(self, task_kwargs):
        return {}

    def get_callback_kwargs_post_document_file_upload(self, task_kwargs):
        return {}


class SourceBackendActionMixinCallbackDocumentUploadBase:
    def get_document_task_kwargs(self, **kwargs):
        result = super().get_document_task_kwargs(**kwargs)

        self._assemble_callback(
            callback_name='post_document_create', result=result,
            task_kwargs=kwargs
        )
        self._assemble_callback(
            callback_name='post_document_file_create', result=result,
            task_kwargs=kwargs
        )
        self._assemble_callback(
            callback_name='post_document_file_upload', result=result,
            task_kwargs=kwargs
        )

        return result


class SourceBackendActionMixinCallbackDocumentFileUploadBase:
    def get_document_file_task_kwargs(self, **kwargs):
        result = super().get_document_task_kwargs(**kwargs)

        self._assemble_callback(
            callback_name='post_document_file_create', result=result,
            task_kwargs=kwargs
        )
        self._assemble_callback(
            callback_name='post_document_file_upload', result=result,
            task_kwargs=kwargs
        )
        return result


class SourceBackendActionMixinCallbackDocumentUpload(
    SourceBackendActionMixinCallbackDocumentUploadBase,
    SourceBackendActionMixinCallbackDocumentFileUploadBase,
    SourceBackendActionMixinCallbackBase
):
    """
    Base callback class for sources that upload documents.
    """


class SourceBackendActionMixinCallbackDocumentFileUpload(
    SourceBackendActionMixinCallbackDocumentFileUploadBase,
    SourceBackendActionMixinCallbackBase
):
    """
    Base callback class for sources that upload document files.
    """
